# -*- coding: utf-8 -*-
from .base import HWSystem
from .base import Wire
from .base import Logic
from .base import Interface

def printHierarchy(sys:Logic):
    """
    Prints the hierarchy of the object

    Parameters
    ----------
    sys : HWSystem
        DESCRIPTION.

    Returns
    -------
    None.

    """
    global indent 
    
    indent = 0;
    
    _printElement(sys)
    
    
def _printElement(obj:Logic):
    
    global indent
    
    indStr = ' ' * indent
    
    print( indStr + type(obj).__name__)
    
    if (len(obj.children) > 0):
        oldindent = indent;
        
        
        for child in obj.children.values():
            indent = oldindent + 1
            _printElement(child)
        
        
        #for child in obj.children:
        #    printElement(child)
    
        indent = oldindent 


def printHierarchyWithValues(sys:Logic, include=None, format=None):
    
    global indent 
    
    indent = 0;
    
    if (format is None):
        format = '{}'
        
    _printElementWithValues(sys, include=include, format=format)
    
    
def _printElementWithValues(obj:Logic, include, format):
    
    global indent
    
    indStr = ' ' * indent

    doPrint = True    
    
    if (not(include is None)):
        # something to filter
        if (not(type(obj).__name__ in include)):
            doPrint = False

    if (doPrint):
        # no filter
        print( indStr + type(obj).__name__ , obj.name, _getPortValues(obj, sfmt=format))
    
    if (len(obj.children) > 0):
        oldindent = indent;
        
        for child in obj.children.values():
            indent = oldindent + 1
            _printElementWithValues(child, include=include, format=format)
            
        indent = oldindent
    
def getPorts(obj:Logic):
    ret = []
    for ip in obj.inPorts:
        ret.append(ip)

    for ip in obj.outPorts:
        ret.append(ip)

    return ret

def getPortWires(obj:Logic):
    ret = []
    for ip in obj.inPorts:
        ret.append(ip.wire)

    for ip in obj.outPorts:
        ret.append(ip.wire)

    return ret

def _getPortValues(obj:Logic, sfmt:str):
    ret = '('
    link = ''
    for ip in obj.inPorts:
        ret = ret + link + ip.name + '=' + sfmt.format(ip.wire.get()) 
        link = ','
        
    ret = ret + ') -> ('
    
    link = ''
    for ip in obj.outPorts:
        ret = ret + link + ip.name + '=' + sfmt.format(ip.wire.get()) 
        link = ','
        
    ret = ret + ') '
    
    return ret

def dumpInterfaceSources(interface:Interface):
    print('Interface name:', interface.name)
    for sourceInfo in interface.sourceToSink:
        name = sourceInfo[0]
        wire = sourceInfo[1]
        print(' Wire: ->' , name)
        
        sourcePort = wire.getSource()
        portParent = sourcePort.parent
        portName = sourcePort.name
        print('  Source Port Parent:', portParent.getFullPath())
        print('  Source Port Name:', portName)
        
    for sinkInfo in interface.sinkToSource:
        name = sinkInfo[0]
        wire = sinkInfo[1]
        print(' Wire: <-' , name)
        
        sourcePort = wire.getSource()
        portParent = sourcePort.parent
        portName = sourcePort.name
        print('  Source Port Parent:', portParent.getFullPath())
        print('  Source Port Name:', portName)
        
        
    
def checkPortParent(port, obj:Logic):
    if (port.parent != obj):
        print('ERROR', obj.name + '['+ port.name+']', ' parent is ',  port.parent.getFullPath(), 'expecting', obj.getFullPath())

def checkPort(port):
    parent = port.parent # parent logic
    
    if (not(port in parent.inPorts or port in parent.outPorts)):
        raise Exception('ERROR: {} not port of parent {}'.format(port.name, parent.getFullPath()) )
    
def checkIntegrity(obj:Logic):
    """
    Checks that the circuit has integrity.
    All wires are correctly connected

    Parameters
    ----------
    obj : Logic
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
    # check that the parents of ports are referering to the same entity
    for inP in obj.inPorts:
        checkPortParent(inP, obj)

    for outP in obj.inPorts:
        checkPortParent(outP, obj)
            
    # check that wires of ports are connected
    for inP in obj.inPorts:
        wire = inP.wire
        sinks = wire.getSinks()
        source = wire.getSource() # gets the source port
        
        if (len(sinks) == 0):
            print('WARNING', obj.getFullPath(), wire.name, 'with no sinks')
        if (source == None):
            raise Exception('ERROR: {} {} with no source'.format(obj.getFullPath(), wire.name))
            
        checkPort(source)
            
    for outP in obj.outPorts:
        wire = outP.wire
        sinks = wire.getSinks()
        source = wire.getSource()
        
        if (len(sinks) == 0):
            print('WARNING', obj.getFullPath(), wire.name, 'with no sinks')
        if (source == None):
            raise Exception('ERROR: {} {} with no source'.format(obj.getFullPath(), wire.name))
            
    for child in obj.children.values():
        checkIntegrity(child)


# -*- coding: utf-8 -*-
"""
Created on Sat Jan 29 16:28:57 2022

@author: dcr
"""

from .base import Wire
from .logic.arithmetic import *
from .logic.bitwise import *
from .logic.simulation import *
from .logic.storage import *
import math

class LogicHelper:
    """
    Helper class to create logic in a simpler way, 
    basically automatically creating the required output wires
    """
    
    def __init__(self, parent:Logic):
        self.parent = parent
        self.inum = 0   # instance number
        
    def hw_abs(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Abs(self.parent, name, a, r)
        return r

    def hw_add(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)+1
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Add(self.parent, name, a, b, r)
        return r

    def hw_signed_add(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Add(self.parent, name, a, b, r)
        return r
    
    def hw_and2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And2(self.parent, name, a, b, r)
        return r

    def hw_and3(self, a:Wire, b:Wire, c:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, [a, b, c], r)
        return r

    def hw_and4(self, a:Wire, b:Wire, c:Wire, d:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, [a, b, c, d], r)
        return r

    def hw_and(self, a:list) -> Wire:
        w = a[0].getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, a, r)
        return r

    def hw_bit(self, a:Wire, bit:int) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name)
        Bit(self.parent, name, a, bit, r)
        return r

    def hw_bits_msbf(self, a:Wire) -> list:
        w = a.getWidth()
        name = self._getNewName() + '_'
        r = self.parent.wires(name, w, 1)
        BitsMSBF(self.parent, name, a, r)
        return r
        
    def hw_bits_lsbf(self, a:Wire) -> list:
        w = a.getWidth()
        name = self._getNewName() + '_'
        r = self.parent.wires(name, w, 1)
        BitsLSBF(self.parent, name, a, r)
        return r

    def hw_reg(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Reg(self.parent, name, d=a, q=r)
        return r

    def hw_buf(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Buf(self.parent, name, a, r)
        return r

    def hw_concatenate_msbf(self, ins:list) -> Wire:
        name = self._getNewName()
        w = 0
        for inw in ins:
            w += inw.getWidth()
            
        r = self.parent.wire(name, w)
        ConcatenateMSBF(self.parent, name, ins, r)
        return r

    def hw_concatenate_lsbf(self, ins:list) -> Wire:
        name = self._getNewName()
        w = 0
        for inw in ins:
            w += inw.getWidth()
            
        r = self.parent.wire(name, w)
        ConcatenateLSBF(self.parent, name, ins, r)
        return r
    
    def hw_constant(self, width:int, v:int) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, width)
        Constant(self.parent, name, v, r)
        return r

    def hw_mod(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Mod(self.parent, name, a, b, r)
        return r

    def hw_div(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Div(self.parent, name, a, b, r)
        return r
    
    def hw_signed_div(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        SignedDiv(self.parent, name, a, b, r)
        return r    

    def hw_equal(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        Equal(self.parent, name, a, b, r)
        return r
 
    def hw_equal_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        EqualConstant(self.parent, name, a, b, r)
        return r
            
    def hw_not_equal_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        NotEqualConstant(self.parent, name, a, b, r)
        return r
    
    def hw_gt_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        return gt
    
    def hw_lt_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        return lt
    
    def hw_ge_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        ge = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        Or2(self.parent, self._getNewName(), gt, eq, ge)
        return ge
    
    def _getNewName(self)->str:
        while (True):
            name = 'i{}'.format(self.inum)
            self.inum += 1
            if not(name in self.parent.children.keys()):
                return name
    
    
    def hw_or2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Or2(self.parent, name, a, b, r)
        return r

    def hw_or(self, a:list) -> Wire:
        w = a[0].getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Or(self.parent, name, a, r)
        return r

    def hw_or_bits(self, a) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name)
        OrBits(self.parent, name, a, r)
        return r

    def hw_range(self, a:Wire, up:int, down:int) -> Wire:
        w = up - down + 1
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Range(self.parent, name, a, up, down, r)
        return r
        
    def hw_xor2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Xor2(self.parent, name, a, b, r)
        return r

    
    def hw_mul(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = wa+wb
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Mul(self.parent, name, a, b, r)
        return r
        
    def hw_mux2(self, sel:Wire, s0:Wire, s1:Wire) -> Wire:
        w = s0.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Mux2(self.parent, name, sel, s0, s1, r)
        return r

    def hw_mux(self, sel:Wire, options:list) -> Wire:
        name = self._getNewName()
        w = options[0].getWidth()
        r = self.parent.wire(name, w)
        Mux(self.parent, name, sel, options, r)
        return r

    def hw_neg(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Neg(self.parent, name, a, r)
        return r

    def hw_not(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Not(self.parent, name, a, r)
        return r

    def hw_sub(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Sub(self.parent, name, a, b, r)
        return r

    def hw_sign(self, a:Wire) -> Wire:
        w = 1
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Sign(self.parent, name, a, r)
        return r
    
    def sign_extend(self, a:Wire, w) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, w)
        SignExtend(self.parent, name, a, r)
        return r
    
    def zero_extend(self, a:Wire, w) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, w)
        ZeroExtend(self.parent, name, a, r)
        return r
    
    def sim_sequence(self, width:int, seq:list) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, width)
        Sequence(self.parent, name, seq, r)
        return r

class IntegerHelper:
    @staticmethod
    def sign(v):
        if (v < 0):
            return -1
        else:
            return 1
        
    @staticmethod
    def signed_to_c2(v, w):
        mask = (1<<w)-1
        return v & mask

    @staticmethod
    def c2_to_signed(v, w):
        sign = 1 << (w-1)
        if ((v & sign) > 0):
            # negative
            return (v - (1 << w))
        else:
            return v
        
def signExtend(v, w, nw):
    s = (v >> (w-1)) & 1
    ns = 0
    if (s == 1):
        ns = (1 << (nw-w)) - 1
    return v | (ns << w)

class FixedPoint:
    def __init__(self, sign_bit, int_bits, frac_bits, v):
        self.sw = sign_bit
        self.iw = int_bits
        self.fw = frac_bits
        self.v = 0
        
        if (type(v) == int):
            self.intToFixedPoint(v)
        elif (type(v) == float):
            self.floatToFixedPoint(v)
        else:
            raise Exception('Unsupported type {}'.format(type(v)))
        
    def getWidths(self):
        return (self.sw, self.iw, self.fw)
        
    def intToFixedPoint(self, v):
        if (v < 0 and self.sw == 0):
            raise Exception('negative values not supported')
            
        w = self.sw + self.iw + self.fw
        
        maxv = 1 << (self.iw -1)
        
        if (v > maxv):
            raise Exception('Value {} greater than max value: {}'.format(v, maxv))
        
        self.v = (v << self.fw) & ((1<<w)-1)
            
    def floatToFixedPoint(self, v):
        if (v < 0 and self.sw == 0):
            raise Exception('negative values not supported')
        
        w = self.sw + self.iw + self.fw
        self.v = int(v * (1 << self.fw)) & ((1<<w)-1)
        
    def add(self, b):
        if (not(isinstance(b, FixedPoint))):
            b = FixedPoint(self.sw, self.iw, self.fw, b)
            
        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        r.v = (self.v + b.v) & ((1<<w)-1)
        return r

    def sub(self, b):
        if (not(isinstance(b, FixedPoint))):
            b = FixedPoint(self.sw, self.iw, self.fw, b)

        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        r.v = (self.v - b.v) & ((1<<w)-1)
        return r

    def mult(self, b):
        if (not(isinstance(b, FixedPoint))):
            oldb = b
            b = FixedPoint(self.sw, self.iw, self.fw, b)
            #print('old b is ', oldb, 'new b is {:08X}'.format( b.v))
    
        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        av = signExtend(self.v, w, w*2)
        bv = signExtend(b.v, w, w*2)
        r.v = ((av * bv)>>self.fw)  & ((1<<w)-1)
        return r

    def toFloatingPoint(self):
        if (((self.v >> (self.iw+self.fw)) & 1) == 1):
            w = self.sw + self.iw + self.fw
            mask = ((1<<w)-1)
            v = ((self.v ^ mask) + 1) & mask
            return -v / (1<<self.fw)
        return self.v / (1<<self.fw)
        
    def copy():
        return FixedPoint(self.sw, self.iw, self.fw , self.v)
    
    @staticmethod
    def fromRawValue(sw, iw, fw, v):
        r = FixedPoint(sw,iw,fw,0)
        r.v = v
        return r
    
class FixedPointHelper:
    pass

class FloatingPointHelper:

    @staticmethod
    def min(a, b):
        if (a == 0.0 and b == 0.0):
            sa = math.copysign(1, a)
            sb = math.copysign(1, b)
            
            if (sa < sb):
                return a
            else:
                return b
        return min(a, b)

    @staticmethod
    def max(a, b):
        if (a == 0.0 and b == 0.0):
            sa = math.copysign(1, a)
            sb = math.copysign(1, b)
            
            if (sa > sb):
                return a
            else:
                return b
        return max(a, b)
    
    @staticmethod
    def fp_to_parts(v):
        if (math.isinf(v) or math.isnan(v)):
            raise Exception('invalid value {}'.format(v))
        e = 0
        s=0
        m = v
        if (m < 0):
            # if v is negative activate the sign, and change the value
            s = 1
            m = -m
    
        if (m > 0.0):
            while (m >= 2):
                # iterate until v is < 2 to determine the exponent
                m = m / 2
                e += 1
                #print('e:', e, 'v:', v)
            while (m < 1):
                m = m * 2
                e -= 1
                #print('e:', e, 'v:', v)
            
        return s, e, m

    @staticmethod
    def sp_to_fixed_point_parts(v):
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            return 0,0,0
        else:
            re = e
            rm = int(round(m * (1<<23)))

            return s, re, rm
    
    @staticmethod
    def sp_to_ieee754_parts(v):
        """
        Return the parts of the IEEE 754 representation of v

        Parameters
        ----------
        v : TYPE
            DESCRIPTION.

        Returns
        -------
        int
            sign.
        int
            biased representation of exponent (e+127)
        int
            representation of mantissa int((m-1)<<23) 

        """
        if (math.isinf(v)):
            s = 0 if v > 0 else 1
            e = 255
            m = 0
            return s,e,m
        
        if (math.isnan(v)):
            s = 0 
            e = 255
            m = (1<<23)-1
            return s,e,m
        
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            return 0,0,0
        else:
            if (e >= 128):
                return s,255,0  # infinity
                
            if (e <= -127):
                # denormalized values
                div = math.pow(2, (-127-e))
                #print('e',e,'div', div)
                m = m / div
                re = 0
                rm = int(round(m * (1 << 22)))
            else:
                im = int(round((m-1) * (1<<23)))
                if (im >= (1<<23)):
                    e += 1
                    m /= 2
                    
                re = 127 + e
                rm = int(round((m-1) * (1<<23)))
    
            return s, re, rm

    def dp_to_ieee754_parts(v):
        """
        Return the parts of the IEEE 754 representation of v

        Parameters
        ----------
        v : TYPE
            DESCRIPTION.

        Returns
        -------
        int
            sign.
        int
            biased representation of exponent (e+1023)
        int
            representation of mantissa int((m-1)<<52) 

        """
        if (math.isinf(v)):
            s = 0 if v > 0 else 1
            e = 2047
            m = 0
            return s,e,m

        if (math.isnan(v)):
            s = 0 
            e = 2047
            m = (1<<51)-1
            return s,e,m

            
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            v = math.copysign(1, v)
            s = 0 if v > 0 else 1
            return s,0,0
        else:
            if (e >= 1024):
                return s,2047,0  # infinity
                
            if (e <= -1023):
                # denormalized values
                div = math.pow(2, (-1023-e))
                #print('e',e,'div', div)
                m = m / div
                re = 0
                rm = int(round(m * (1 << 51)))
            else:
                im = int(round((m-1) * (1<<52)))
                if (im >= (1<<52)):
                    e += 1
                    m /= 2
                    
                re = 1023 + e
                rm = int(round((m-1) * (1<<52)))
    
            return s, re, rm

    @staticmethod
    def sp_to_ieee754(v):
        s,e,m = FloatingPointHelper.sp_to_ieee754_parts(v)
        
        r = s << 31
        r = r | (e << 23)
        r = r | (m)
        return r
    
    @staticmethod
    def dp_to_ieee754(v):
        s,e,m = FloatingPointHelper.dp_to_ieee754_parts(v)
        
        r = s << 63
        r = r | (e << 52)
        r = r | (m)
        return r

    @staticmethod
    def parts_to_fp(s, e, m):
        return math.pow(-1, s) * math.pow(2, e) * m

    @staticmethod
    def ieee754_parts_to_sp(s, e, m):
        """
        We build a floating point number from is sign/exponent/mantisa representation
        as it is store in the IEEE754 format

        Parameters
        ----------
        s : int
            sign.
        e : int
            exponent.
        m : TYPE
            mantisa.

        Returns
        -------
        float
            floating point number r = (-1)^s * 2^e * m

        """
        
        # zero is a special case
        if (e == 0 and m == 0):
            if (s == 1):
                return -0.0
            else:
                return 0.0
            
        if (e == 0):
            ef = -126
            mf = m / (1<<23)
        else:
            ef = e-127
            mf = ((1 << 23) | m) / (1<<23)
        
        return FloatingPointHelper.parts_to_fp(s , ef , mf)
    
    @staticmethod
    def ieee754_parts_to_dp(s, e, m):
        """
        We build a floating point number from is sign/exponent/mantisa representation
        as it is store in the IEEE754 format

        Parameters
        ----------
        s : int
            sign.
        e : int
            exponent.
        m : TYPE
            mantisa.

        Returns
        -------
        float
            floating point number r = (-1)^s * 2^e * m

        """
        
        # zero is a special case
        if (e == 0 and m == 0):
            if (s == 1):
                return -0.0
            else:
                return 0.0
            
        if (e == 0):
            ef = -1022
            mf = m / (1<<52)
        else:
            ef = e-1023
            mf = ((1 << 52) | m) / (1<<52)
        
        return FloatingPointHelper.parts_to_fp(s , ef , mf)
    
    @staticmethod
    def ieee754_to_sp(v):
        if (v == 0):
            return 0.0
        
        s = v >> 31
        e = (v >> 23) & 0xFF
        m = (v & ((1<<23)-1))  
        
        if (e == 255):
            if (m == 0):
                return -math.inf if (s == 1) else math.inf
            else:
                return math.nan
        
        return FloatingPointHelper.ieee754_parts_to_sp(s, e, m)

    @staticmethod
    def ieee754_to_dp(v):
        if (v == 0):
            return 0.0
        
        s = v >> 63
        e = (v >> 52) & 0x7FF
        m = (v & ((1<<52)-1))  
        
        if (e == 0x7FF):
            if (m == 0):
                return -math.inf if (s == 1) else math.inf 
            else:
                return math.nan

        return FloatingPointHelper.ieee754_parts_to_dp(s, e, m)

    @staticmethod
    def ieee754_stored_internally(v):
        s,e,m = FloatingPointHelper.sp_to_ieee754_parts(v)
        iv = FloatingPointHelper.ieee754_parts_to_sp(s, e, m)
        return iv

    @staticmethod
    def zp_bin(v:int, vl:int) -> str:
        """        
        Generates a the binary representation of an integer value in a number
        of bits. Leading zeros are added as needed.
        
        Parameters
        ----------
        v : int
            value.
        vl : int
            length of the output.

        Returns
        -------
        TYPE
            the binary string.

        """
        if (v >= 0):
            str = bin(v)[2:]
        else:
            # bin function does not handle 2's complement,
            # so let's do it for them
            p = -v
            mask = (1 << vl) -1
            r = (mask - p + 1) & mask
            str = bin(r)[2:] 
            
        return ('0' * (vl-len(str))) + str
    
    
class BitManipulation:
    @staticmethod
    def getFirstBit(v:int) -> int:
        # returns the position of the first active bit
        n = 0
        while (True):
            if (v == 0):
                return n
            n += 1
            v = v >> 1
    
    @staticmethod
    def countLeadingZeros(v:int, w:int) -> int:
        return w - BitManipulation.getFirstBit(v)

    @staticmethod
    def getBit(v:int, bit:int) -> int:
        return (v >> bit) & 0x1
    
class CircuitAnalysis:

    @staticmethod
    def getAllPorts(obj:Logic) -> list:
        ret = []
        
        ret.extend(obj.inPorts)
        ret.extend(obj.outPorts)
            
        return ret
    
    def getAllWireNames(obj:Logic) -> list:
        """
        Return the list of wire names

        Parameters
        ----------
        obj : Logic
            DESCRIPTION.

        Returns
        -------
        None.

        """
        return [x for x in obj._wires.keys()]
    
    @staticmethod
    def getAllPortNames(obj:Logic) -> list:
        """
        Get all port names of an object

        Parameters
        ----------
        obj : Logic
            DESCRIPTION.

        Returns
        -------
        list
            DESCRIPTION.

        """
        ret = []
        
        for p in obj.inPorts:
            ret.append(p.name)
            
        for p in obj.outPorts:
            ret.append(p.name)
            
        return ret

    @staticmethod
    def has_clk_enable(obj:Logic) -> bool:
        input_port_names = [p.name for p in obj.inPorts]
        return 'clk_enable' in input_port_names
        
    @staticmethod
    def getAllPortWires(obj:Logic) -> list:
        """
        Returns a list of the wires connected to all the ports of a circuit

        Parameters
        ----------
        obj : Logic
            the object to analyze.

        Returns
        -------
        list
            list with all the wires connected to the circuit ports.

        """
        ret = []
        
        for p in obj.inPorts:
            ret.append(p.wire)
            
        for p in obj.outPorts:
            ret.append(p.wire)
            
        return ret


# -*- coding: utf-8 -*-
"""
Created on Sun Apr  2 13:59:30 2023

@author: dcr
"""

from .. import *
from deprecated import deprecated

import ast
from .astutils import * 

def startsWith(line, sub):
    if (line[0:len(sub)] == sub):
        return True
    else:
        return False
    
def strip(line):
    line = line.strip()
    
    while (startsWith(line, '\t')):
        # remove tabs
        line = line[1:]

    return line

def getBody(node, slist=''):
    # the Module node contains a body, that contains a list, containting
    # a function definition with a body
    
    # AST visitors can not deal directly with list, we wrap them in 
    # a dummy verilog body object
    var = VerilogDeclarations()
    init = VerilogInitial()
    process = VerilogProcess(node.body[0].body, slist)
    return VerilogBody(var, init, process)
    
class Python2VerilogTranspiler:

    def __init__(self, obj:Logic):
        self.obj = obj
        self.signals = {}
        self.indent = 0

    def getIndent(self):
        return ' ' * (self.indent * 4)


    def transpileCombinational(self):
        '''
        Transpile RTL style behavioural descriptions

        Returns
        -------
        str
            the equivalent RTL

        '''
        module = getMethod(self.obj, 'propagate')
    
        node = getBody(module, '*')
        
        initExtracter = ExtractInitializers()
        init = initExtracter.visit(node)

    
        assert(isinstance(node, ast.AST))
        
        node = RemovePrints().visit(node)
        node = RemoveAssert().visit(node)
        
        node = ReplaceIf().visit(node)
        node = ReplaceWireCalls().visit(node)
        node = ReplaceExpr().visit(node)
        node = ReplaceOperators().visit(node)
        node = ReplaceOperators().visit(node) # repeat to handle Compare

        #node = ReplaceAttribute().visit(node)
        wiresAndVars = ReplaceWiresAndVariables(initExtracter.ports, initExtracter.variables)
        node = wiresAndVars.visit(node)
        node = ReplaceConstant().visit(node)
        node = ReplaceAssign().visit(node)
 
        node = FlattenOperators().visit(node)

        return node
        #return Python2VerilogTranspiler.toVerilog(node)
        #return self.transpileUnknown(node)

    def transpileSequential(self):
        '''
        Transpile RTL style behavioural descriptions

        Returns
        -------
        str
            the equivalent RTL

        '''
        # start analyzing the constructor to get wires and variables
        module = getMethod(self.obj, '__init__')
        node = getBody(module)
        
        initExtracter = ExtractInitializers()
        init = initExtracter.visit(node)
        
        module = getMethod(self.obj, 'clock')
        clkname = getObjectClockDriver(self.obj).name

        node = getBody(module, 'posedge {}'.format(clkname))
        node.init.body = init.init.body
    
        assert(isinstance(node, ast.AST))
        
        node = RemovePrints().visit(node)
        node = RemoveAssert().visit(node)
        
        node = ReplaceIf().visit(node)
        node = ReplaceWireCalls().visit(node)
        node = ReplaceExpr().visit(node)
        node = ReplaceOperators().visit(node)
        node = ReplaceOperators().visit(node) # repeat to handle Compare
        node = ReplaceOperators().visit(node) # repeat to handle Compare
        
        wiresAndVars = ReplaceWiresAndVariables(initExtracter.ports, initExtracter.variables)
        node = wiresAndVars.visit(node)
        node = ReplaceConstant().visit(node)
        node = ReplaceAssign().visit(node)
        node = ReplaceIfExp().visit(node)
        node = ReplaceDocStrings().visit(node)
        
        node.wires.variables = wiresAndVars.variables.values();
        #node = FlattenOperators().visit(node)

        # print('variables', node.wires.variables)
        
        return node;
        # return Python2VerilogTranspiler.toVerilog(node)
        #return self.transpileUnknown(node)
        
    @staticmethod
    def toVerilog(node):
        '''
        Tranlates to Verilog a node by calling its toVerilog method
        For lists, in invokes the same method to the elements of the list

        Parameters
        ----------
        node : TYPE
            DESCRIPTION.

        Returns
        -------
        toV : TYPE
            DESCRIPTION.

        '''
        #print('transpiling', type(node))

        if (isinstance(node, list)):
            str = ''
            for item in node:
                str += Python2VerilogTranspiler.toVerilog(item)
            return str
        
        if not(hasattr(node, 'toVerilog')):
            raise Exception('No toVerilog for', type(node), node)
        else:
            toV = getattr(node, 'toVerilog')
            return toV()

    def format(self, str):
        lines = str.split('\n')
        indent = 0
        ret = ''
        
        for line in lines:
            line = strip(line)
            
            if (startsWith(line, 'end')):
                indent -= 1
                #print('end->indent: ', indent)
                
            sindent = ''
            if (indent > 0):
                sindent = ' ' * (indent * 4)
                #print('len indent:', len(sindent))
                #ret += '{:02}-'.format(indent)  + sindent
                
            if (startsWith(line, 'begin')):
                indent += 1
                #print('begin->indent: ', indent)
                
            if (len(line) > 0):
                ret += sindent  + line + '\n'
        
        return ret
    
    @deprecated
    def getExtraDeclarations(self):
        from py4hw.rtl_generation import getValidVerilogName

        str = ""
        
        portNames = []
        
        
        for inp in self.obj.inPorts:
            portNames.append(getValidVerilogName(inp.name))

        for outp in self.obj.outPorts:
            portNames.append(getValidVerilogName(outp.name))        
        
        #print('SIGNALS:', self.signals)
        #print('PORT NAMES:', portNames)

        extra = [x for x in self.signals if x not in portNames]
        
        # @todo we should analyze the number of possible values of 
        # extra signals to decide their width. By now we consider a worst
        # case scenario with extra signals all requiring a maximum of 8 bits
        # this will be simplyfied during synthesis if less bits are required
        # but it will cause a BUG if the required bits are higher
        for sig in extra:
            str += "reg [7:0] " + sig + " = 0;\n" 
            
        return str
    
    
    
    
class ReplaceIf(ast.NodeTransformer):
        
    def visit_If(self, node):
        #print('transpiling if')
        condition = node.test
        
        positive = node.body
        negative = node.orelse
        
        node2 = VerilogIf(condition, positive, negative)

        node2 = ast.NodeTransformer.generic_visit(self, node2)
        return node2
        

class RemovePrints(ast.NodeTransformer):
        
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        #print('checking call', attr)
        if (attr == 'print'):
            # remove prints
            return VerilogComment('print removed')
        
        node = ast.NodeTransformer.generic_visit(self, node)
        
        return node

class RemoveAssert(ast.NodeTransformer):
        
    def visit_Assert(self, node):
        from py4hw.rtl_generation import getAstName

        # remove asserts
        return VerilogComment('assert removed')
        
class ReplaceDocStrings(ast.NodeTransformer):
    
    def visit_VerilogProcess(self, node):
        newbody = []
        
        for obj in node.body:
            if (isinstance(obj, VerilogConstant)):
                obj = VerilogComment(obj.value)
            newbody.append(obj)
        return VerilogProcess(newbody, node.sensitivity_list)
    
class ReplaceWireCalls(ast.NodeTransformer):
        
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        #print('checking call', attr)
        if (attr == 'get'):
            #if isinstance(node.func.value, ast.Attribute):
            wirename = getAstName(node.func.value)
            #print('REPLACING GET ', wirename)
            return VerilogWire(wirename)
        
        elif (attr == 'prepare'):
            #print('REPLACE WIRE PUTS FUNC:', attr , node.func.value.attr, node.args)
            left = VerilogWire(node.func.value.attr)
            node = VerilogSynchronousAssignment(left, node.args[0])
        
        elif (attr == 'put'):
            #print('REPLACE WIRE PUTS FUNC:', attr , node.func.value.attr, node.args)
            left = VerilogWire(node.func.value.attr)
            right = ast.NodeTransformer.generic_visit(self, node.args[0])
            node = VerilogAsynchronousAssignment(left, right)
        else:
            print('WARNING: unhandled call {}'.format(attr))
                 
        node = ast.NodeTransformer.generic_visit(self, node)
        
        return node
    
    

class ReplaceOperators(ast.NodeTransformer):
    
    def visit_BinOp(self, node):
        #print('replacing BinOp')
        left = ast.NodeTransformer.generic_visit(self, node.left)
        right = ast.NodeTransformer.generic_visit(self, node.right)
        node =  VerilogOperator(left, node.op, right)
        #node = ast.NodeTransformer.generic_visit(self, node)
        return node

    def visit_BoolOp(self, node):
        assert(len(node.values) == 2)
        left = ast.NodeTransformer.generic_visit(self, node.values[0])
        right = ast.NodeTransformer.generic_visit(self, node.values[1])
        node = VerilogOperator(left, node.op, right)
        return node
    
    def visit_Compare(self, node):
        # print('visiting compare')
        left = ast.NodeTransformer.generic_visit(self, node.left)
        right = node.comparators # ast.NodeTransformer.generic_visit(self, node.comparators)
        node = VerilogOperator(left, node.ops, right)            
        return node
    

    
class ReplaceExpr(ast.NodeTransformer):
    def visit_Expr(self, node):
        
        return node.value

class ReplaceIfExp(ast.NodeTransformer):
    def visit_IfExp(self, node):
        cond = ast.NodeTransformer.generic_visit(self, node.test)
        positive = ast.NodeTransformer.generic_visit(self, node.body)
        negative = ast.NodeTransformer.generic_visit(self, node.orelse)
        return VerilogTernaryConditionalOperator(cond, positive, negative)
        

class ReplaceWiresAndVariables(ast.NodeTransformer):
    def __init__(self, ports, variables):
        self.ports = ports
        self.variables = variables
        
    def visit_Name(self, node):
        name = node.id
        if (name in self.ports.keys()):
            return VerilogWire(name)

        if (name in self.variables.keys()):
            return VerilogVariable(name, self.variables[name].type)

        # create new variable
        self.variables[name] = VerilogVariableDeclaration(name, 'integer')
        return VerilogVariable(name, 'integer')

    def visit_Attribute(self, node):
        name = node.attr
        if (name in self.ports.keys()):
            return VerilogWire(name)

        if (name in self.variables.keys()):
            return VerilogVariable(name, self.variables[name].type)

        # create new variable
        self.variables[name] = VerilogVariableDeclaration(name, 'integer')
        return VerilogVariable(name, 'integer')
    
class ReplaceConstant(ast.NodeTransformer):
    def visit_Constant(self, node):
        
        return VerilogConstant(node.value)
    
    def visit_Num(self, node):
        return VerilogConstant(node.n)
    
class ReplaceAssign(ast.NodeTransformer):
            
    def visit_Assign(self, node):
        if (len(node.targets) > 1):
            raise Exception('{} targets! only 1 is supported'.format(len(node.targets)))
            
        left = ast.NodeTransformer.generic_visit(self, node.targets[0])
        right = ast.NodeTransformer.generic_visit(self, node.value)
            
        if (isinstance(left, VerilogVariable)):
            node = VerilogVariableAssignment(left, right)
        else:
            node =  VerilogSynchronousAssignment(left, right)
        
        return node
    
    def visit_AugAssign(self, node):
        left = node.target
        right = node.value
        
        newvalue = VerilogOperator(left, node.op, right)
        if (isinstance(left, VerilogVariable)):
            node = VerilogVariableAssignment(left, newvalue)
        else:
            node = VerilogSynchronousAssignment(left, newvalue)
        
        return node

class ExtractInitializers(ast.NodeTransformer):
    ports = {}
    variables = {}
    
    # Extracts initializers from class constructor
    def visit_Constant(self, node):
        return VerilogConstant(node.value)
    
    def visit_Num(self, node):
        return VerilogConstant(node.n)

    def visit_VerilogBody(self, node):
        self.top = node
        return ast.NodeTransformer.generic_visit(self, node)
    
    def visit_Assign(self, node):
        from py4hw.rtl_generation import getAstName

        if (isinstance(node.value, ast.Call)):
            fname = getAstName(node.value.func)
            pname = ''
            
            if (fname == 'addIn'):
                pname = node.targets[0].attr
                # print('in port', pname)
            elif (fname == 'addOut'):
                node.targets[0].attr
                # print('out port', node.targets[0].attr)
            else:
                print('name not expected', fname)

            w = VerilogWire(pname)
            self.ports[pname] = w        
            self.top.wires.wires.append(w)
            return None
        
        elif (isinstance(node.value, ast.Num)):
            vname = node.targets[0].attr
            var = VerilogVariable(vname, type(node.value.n))
            self.top.wires.variables.append(var)
            node = VerilogVariableAssignment(var, node.value)
            node = ast.NodeTransformer.generic_visit(self, node)
            self.top.init.body.append(node)

            # save variable
            self.variables[vname] = VerilogVariableDeclaration(vname, 'integer')
            return None
        
        else:
            raise Exception('not handled')
        
        return node
                
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        #print('checking call', attr)
        if (attr == '__init__'):
            return None
        
        return node

class FlattenOperators(ast.NodeTransformer):
    # If recursive operators are found they are extracted, new wires
    # are created and the structure is flattened
    ic = -1

    def loop_visit(self, node):
        self.anyChange = True
        
        while (self.anyChange):
            self.anyChange = False
            node = self.visit(node)
            
        return node
            
    def visit_VerilogBody(self, node):
        self.top = node
        return ast.NodeTransformer.generic_visit(self, node)
    
    def newName(self):
        self.ic += 1
        return 'i{}'.format(self.ic)
    
    def visit_VerilogAsynchronousAssignment(self, node):
        self.sync = False
        return ast.NodeTransformer.generic_visit(self, node)

    def visit_VerilogSynchronousAssignment(self, node):
        self.sync = True
        return ast.NodeTransformer.generic_visit(self, node)
        
    def visit_VerilogOperator(self, node):
        if (isinstance(node.left, VerilogOperator)):
            #print('we should extract left operator')
            wn = self.newName()
            vwd = VerilogWireDeclaration(wn)
            vw = VerilogWire(wn)
            self.top.wires.wires.append(vwd)
            
            if (self.sync):
                assign = VerilogSynchronousAssignment(vw, node.left)
            else:
                assign = VerilogAsynchronousAssignment(vw, node.left)
                
            self.top.process.body.append(assign)
            node.left = vw
            self.anyChange = True 
            
        if (isinstance(node.right, VerilogOperator)):
            #print('we should extract right operator')
            wn = self.newName()
            vwd = VerilogWireDeclaration(wn)
            vw = VerilogWire(wn)
            self.top.wires.wires.append(vwd)
            
            if (self.sync):
                assign = VerilogSynchronousAssignment(vw, node.right)
            else:
                assign = VerilogAsynchronousAssignment(vw, node.right)
                
            self.top.process.body.append(assign)
            node.right = vw
            self.anyChange = True
            
        return node


class VerilogVariableAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) + '='
        str += Python2VerilogTranspiler.toVerilog(self.right) + ';\n'
        return str
        
class VerilogAsynchronousAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) + '<='
        str += Python2VerilogTranspiler.toVerilog(self.right) + ';\n'
        return str
        
class VerilogSynchronousAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) + '<='
        str += Python2VerilogTranspiler.toVerilog(self.right) + ';\n'
        return str

class VerilogOperator(ast.AST):
    def __init__(self, left, op, right):
        if (isinstance(op, list)):
            # we only support single operators
            assert(len(op) == 1)
            op = op[0]
            
        self.left = left
        self.op = self.getOp(op)
        self.right = right
        self._fields = tuple(['left', 'op', 'right'])

    def getOp(self, operator):
        # translated an AST operator into verilog syntax
        # ARITHMETIC
        if (isinstance(operator, ast.Add)):
            return '+'
        elif (isinstance(operator, ast.Sub)):
            return '-'
        elif (isinstance(operator, ast.Mult)):
            return '*'
        elif (isinstance(operator, ast.FloorDiv)):
            return '/'
        # BITWISE
        elif (isinstance(operator, ast.BitAnd)):
            return '&'
        elif (isinstance(operator, ast.BitOr)):
            return '|'
        elif (isinstance(operator, ast.BitXor)):
            return '^'
        elif (isinstance(operator, ast.LShift)):
            return '<<'
        elif (isinstance(operator, ast.RShift)):
            return '>>'
        # RELATIONAL
        elif (isinstance(operator, ast.And)):
            return '&&'
        elif (isinstance(operator, ast.Eq)):
            return '=='
        elif (isinstance(operator, ast.Lt)):
            return '<'        
        elif (isinstance(operator, ast.GtE)):
            return '>='        
        elif (isinstance(operator, ast.Mod)):
            return '%'
        else:
            raise Exception('operator {} not supported'.format(type(operator)))
            
    def toVerilog(self):
        str = ''
        
        if (isinstance(self.left, VerilogOperator)):
            str += '(' + Python2VerilogTranspiler.toVerilog(self.left)  + ')'
        else:
            str += Python2VerilogTranspiler.toVerilog(self.left) 
            
        str += self.op

        if (isinstance(self.right, VerilogOperator)):
            str += '(' + Python2VerilogTranspiler.toVerilog(self.right) + ')'
        else:
            str += Python2VerilogTranspiler.toVerilog(self.right)

        return str


class VerilogWire(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str):
        self.name = name
        self._fields = tuple(['name', 'dummy'])

    def toVerilog(self):
        return self.name

class VerilogVariable(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str, type:str):
        self.name = name
        self.type = type
        self._fields = tuple(['name', 'type'])

    def toVerilog(self):
        return self.name
    
class VerilogWireDeclaration(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str):
        self.name = name
        self._fields = tuple(['name', 'dummy'])

    def toVerilog(self):
        return 'reg ' + self.name + ';\n'


class VerilogVariableDeclaration(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str, vartype:str):
        self.name = name
        self.type = vartype
        self._fields = tuple(['name', 'type'])

    def toVerilog(self):
        return '{} {};\n'.format(self.type, self.name )


class VerilogDeclarations(ast.AST):
    # wire declaration section at the beginning of the module
    def __init__(self):
        self.wires = []
        self.variables = []
        self._fields = tuple(['wires', 'variables'])
        
    def toVerilog(self):
        str = ''
        
        for w in self.wires:
            str += Python2VerilogTranspiler.toVerilog(w)

        for v in self.variables:
            str += Python2VerilogTranspiler.toVerilog(v)
            
        return str    

class VerilogInitial(ast.AST):
    def __init__(self):
        self.body = []
        self._fields = tuple(['body', 'dummy'])
        
    def toVerilog(self):
        str = 'initial\n'
        str += 'begin\n'
        
        for st in self.body:
            str += Python2VerilogTranspiler.toVerilog(st) 

        str += 'end\n'            
        return str            

class VerilogProcess(ast.AST):
    def __init__(self, body, sensitivity):
        self.body = body
        self.sensitivity_list = sensitivity
        self._fields = tuple(['body', 'sensitivity_list'])
        
    def toVerilog(self):
        str = 'always @({})\n'.format(self.sensitivity_list)
        str += 'begin\n'
        
        for st in self.body:
            str += Python2VerilogTranspiler.toVerilog(st) 

        str += 'end\n'            
        return str            
    
class VerilogBody(ast.AST):
    '''
    AST node to wrap body
    '''
    def __init__(self, wires, initial, process):
        self.wires = wires
        self.init = initial
        self.process = process
        self._fields = tuple(['wires', 'init','process'])
    
    def toVerilog(self):
        str = ''
        
        str += '// wire/variable declaration \n'
        str += Python2VerilogTranspiler.toVerilog(self.wires)

        str += '// initial \n'
        str += Python2VerilogTranspiler.toVerilog(self.init)

        str += '// process \n'
        str += Python2VerilogTranspiler.toVerilog(self.process)
        return str
            
            
class VerilogIf(ast.AST):
    def __init__(self, condition, positive, negative):
        self.condition = condition
        self.positive = positive
        self.negative = negative
        self._fields = tuple(['condition', 'positive', 'negative'])
        
    def toVerilog(self):
        str = 'if (' + Python2VerilogTranspiler.toVerilog(self.condition) + ')\n'
        str += 'begin\n'
        str += Python2VerilogTranspiler.toVerilog(self.positive) + '\n'
        str += 'end\n'
        
        if (len(self.negative) > 0):
            str += 'else\n'
            str += 'begin\n'
            str += Python2VerilogTranspiler.toVerilog(self.negative) + '\n'
            str += 'end\n'
        return str
    
class VerilogComment(ast.AST):
    def __init__(self, value):
        self.value = value
        self._fields = tuple(['value', 'dummy'])
    
    def toVerilog(self):
        return '/* {} */\n'.format(self.value)
    
class VerilogConstant(ast.AST):
    def __init__(self, value):
        self.value = value
        self._fields = tuple(['condition', 'positive', 'negative'])
        
    def toVerilog(self):
        return '{}'.format(self.value)
    
class VerilogTernaryConditionalOperator(ast.AST):
    def __init__(self, cond, positive, negative):
        self.condition = cond
        self.positive = positive
        self.negative = negative
        self._fields = tuple(['condition', 'positive', 'negative'])

    def toVerilog(self):
        return '({}) ? {} : {}'.format(Python2VerilogTranspiler.toVerilog(self.condition),
            Python2VerilogTranspiler.toVerilog(self.positive),
            Python2VerilogTranspiler.toVerilog(self.negative))


import tkinter 
import tkinter.font
from tkinter import *
from tkinter import ttk
from PIL import Image, ImageTk
import os
from .base import Logic
from .schematic import Schematic
    
class Workbench():
    
    def getIcon(self, path):
        script_directory = os.path.dirname(os.path.abspath(__file__))
        icon_path = os.path.join(script_directory, path)
        
        icon = Image.open(icon_path)
        icon = ImageTk.PhotoImage(icon)
        return icon
        
    def __init__(self, sys:Logic):
    
        self.sys = sys
        self.detailObj = None
        
        root = tkinter.Tk()
        root.title('py4hw Interactive Workbench')
        
        self.root = root
        
        ttk.Style().configure("Treeview", fg="light yellow")
        font = tkinter.font.Font(size=8)
        ttk.Style().configure("Prolepsis.Treeview", font=font)
        
        self.topPane = PanedWindow(root, orient=HORIZONTAL)
        
        self.hierarchyPane = Frame(self.topPane) # PanedWindow(self.topPane, relief = SUNKEN, width=50, height=100)
        self.topPane.add(self.hierarchyPane)
        
        self.buttonPane = Frame(self.hierarchyPane) #, relief = SUNKEN, width=100, height=100)
        
        self.txtClocks = tkinter.StringVar(value='1')
        #print(ttk.Style().lookup("Prolepsis.Treeview", "font"))
        txtClocks = ttk.Entry(self.buttonPane, width=10, textvariable=self.txtClocks)
        #txtClocks.grid(row=0, column=0)
        txtClocks.pack(side=tkinter.LEFT, padx=10, pady=(10,5))
        
        btnClock = ttk.Button(self.buttonPane, text="Clock", command=self.guiClk)
        btnClock.pack(side=tkinter.RIGHT, padx=(10,0), pady=(10,5))
        #btnClock.grid(row=0, column=1)
        
        #self.hierarchyPane.add(self.buttonPane)
        
        self.buttonPane.pack(side=tkinter.TOP)
        
        self.createHierarchyTree()
        
        self.rightPane = PanedWindow(self.topPane, orient=VERTICAL, relief = SUNKEN, width=100, height=100)
        self.topPane.add(self.rightPane)
        
        self.interfacePane = PanedWindow(self.rightPane, relief = SUNKEN, width=100, height=100)
        
        self.schematicAreaPane = PanedWindow(self.rightPane,  orient=VERTICAL, relief = SUNKEN, width=100, heigh=20)
        self.schematicToolbarPane = Frame(self.schematicAreaPane) #,  orient=HORIZONTAL, relief = SUNKEN, width=100, heigh=20)
                
        self.rightPane.add(self.interfacePane)
        self.rightPane.add(self.schematicAreaPane)
        self.schematicAreaPane.add(self.schematicToolbarPane) 
        
        icon_zo = self.getIcon('zoomout24.png')
        icon_zi = self.getIcon('zoomin24.png')
        
        btnZoomIn = ttk.Button(self.schematicToolbarPane,  image=icon_zi, text="Zoom in", command=self.guiZoomIn)
        btnZoomOut = ttk.Button(self.schematicToolbarPane, image=icon_zo, text="Zoom out", command=self.guiZoomOut)
        
        btnZoomIn.pack(side=tkinter.LEFT)
        btnZoomOut.pack(side=tkinter.LEFT)
        
        self.schematicPane = PanedWindow(self.schematicAreaPane, relief = SUNKEN, width=100, height=100)
        self.schematicAreaPane.add(self.schematicPane)
        
        #pane2.pack(fill=BOTH, expand=YES, side=RIGHT)
        
        self.circuitDetail(None)
        
        self.topPane.pack(fill=BOTH, expand=True)
        
        # explicitelly update
        # sys.getSimulator().addListener(self)
        
        
        root.mainloop()
        
    def guiClk(self):
        clks = int(self.txtClocks.get())
        sim = self.sys.getSimulator()
        sim.clk(clks)
        #explicitelly update
        self.simulatorUpdated()

    def guiZoomOut(self):
        self.schematicDiagram.scale("all", 0, 0, 0.9, 0.9)

    def guiZoomIn(self):
        self.schematicDiagram.scale("all", 0, 0, 1.1, 1.1)

    
    def simulatorUpdated(self):
        """
        TODO: We do the dirty trick to recreate all the table to just
        update the circuit status. We will try to do better in the future

        Returns
        -------
        None.

        """
        if (self.detailObj != None):
            self.setCircuitDetail(self.detailObj)
        
    def circuitDetail(self, obj:Logic):

        # this is called just once now (?)        
        print('Circuit detail on ', obj)
        
        # Create the Circuit Interface pane
        # @todo we are creating this every time we focus on a circuit, could we
        # just create it once and clear the tree view and populate it as necessary ?
        self.detail_tv = ttk.Treeview(self.interfacePane)
        
        detail_tv = self.detail_tv
        self.interfacePane.add(detail_tv)
        
        detail_tv.pack(fill=BOTH, expand=YES)
        
        detail_tv.config(columns=[ 'direction', 'width', 'value'])
        detail_tv.config(displaycolumns=[ 0, 1,2])

        detail_tv.heading('#0', text='Name')
        detail_tv.heading('direction', text='Direction')
        detail_tv.heading('width', text='Width')
        detail_tv.heading("value", text="Value")
        
        detail_tv.column("#0",minwidth=0,width=100)
        detail_tv.column("#0",minwidth=0,width=100)
        detail_tv.column('direction', minwidth=0,width=100)
        detail_tv.column('width', minwidth=0,width=100)
        detail_tv.column("value", minwidth=0,width=100)
        
        detail_tv.config(selectmode=tkinter.NONE)
        detail_tv.tag_configure("ally", background="green")
        #detail_tv.tag_bind("char", "<Double-Button-1>", event)
        #tv.config(style="Prolepsis.Treeview")
        detail_tv["style"] = "Prolepsis.Treeview"

        #self.debugTkinterHierarchy(self.root, 0)
        
        if (obj == None):
            return
        
        self.setCircuitDetail(obj)
        
        
    def debugTkinterHierarchy(self, obj, ii):
        indent = '|' * ii + '+'
        print('[INFO] {} tkinter obj {} {}'.format(indent, type(obj), obj._name if (hasattr(obj, '_name')) else ''));
        
        try:
            for child in obj.children.values():
                self.debugTkinterHierarchy(child, ii+1)
        except:
            pass
                
    def setCircuitDetail(self, obj:Logic):
        
        if (obj != None):
            for pane in self.schematicPane.panes():
                self.schematicPane.forget(pane)
                
            if (obj.isStructural()):
                # ignore non structural circuits
                sch = Schematic(obj, render='tkinter', parent=self.schematicPane)
    
                sch.drawAll()
                self.schematicDiagram = sch.canvas.canvas
                self.schematicPane.add(self.schematicDiagram)
                #self.schematicPane.pack()
            
        self.detailObj = obj
        self.detail_tv.delete(*self.detail_tv.get_children())
        
        detail_tv = self.detail_tv
        
        visitedPorts = []
        for source in obj.sources:
            sWidth = ''
            sValue = ''
            in_id = detail_tv.insert("", tkinter.END, text=source.name, values=[sWidth, sValue], open=True)
            
            for inp in source.ports:
                sWidth = inp.wire.getWidth()
                sValue = hex(inp.wire.get())
                port_id = detail_tv.insert(in_id, tkinter.END, text=inp.name, values=['in', sWidth, sValue], open=True)
                visitedPorts.append(inp)
        
        for sink in obj.sinks:
            sWidth = ''
            sValue = ''
            in_id = detail_tv.insert("", tkinter.END, text=sink.name, values=[sWidth, sValue], open=True)
        
            for inp in sink.ports:
                sWidth = inp.wire.getWidth()
                sValue = hex(inp.wire.get())
                port_id = detail_tv.insert(in_id, tkinter.END, text=inp.name, values=['out', sWidth, sValue], open=True)
                visitedPorts.append(inp)
            
        for inp in obj.inPorts:
            if (inp not in visitedPorts):
                sWidth = inp.wire.getWidth()
                sValue = hex(inp.wire.get())
                in_id = detail_tv.insert("", tkinter.END, text=inp.name, values=['in', sWidth, sValue], open=True)
        
        for outp in obj.outPorts:
            if (outp not in visitedPorts):
                sWidth = outp.wire.getWidth()
                sValue = hex(outp.wire.get())
                in_id = detail_tv.insert("", tkinter.END, text=outp.name, values=['out', sWidth, sValue], open=True)
           
        
        #tv.config(bg="light yellow")
        
    def callback(self, event):
        tvid = self.hierarchyTree.focus()
        obj = self.map_id_obj[tvid]
        # values = tv.item(tvid)['values']
        # objid =
        # obj = sys.getFromFullPath(values[2])
        #sys.getFromFullPath()
        self.setCircuitDetail(obj)
         
    
    def populateTree(self, tv, tvid, obj:Logic):
        for child in obj.children.values():
            instanceName = child.name
            circuitType = 'Structural'
            if (child.isClockable()):
                circuitType = 'Clockable'
            elif (child.isPropagatable()):
                circuitType = 'Propagatable'
            childid = tv.insert(tvid, tkinter.END, text=type(child).__name__, values=[instanceName, circuitType, child.getFullPath()], tags=["ally","char"])
            self.map_id_obj[childid] = child
            self.populateTree(tv, childid, child)

    def createHierarchyTree(self):
        self.hierarchyTree = ttk.Treeview(self.hierarchyPane)
    
        tv = self.hierarchyTree  
        self.map_id_obj = {}
    
        
        tv.bind('<<TreeviewSelect>>', self.callback)
        #self.hierarchyPane.add(tv)
        tv.pack(fill=BOTH, expand=YES)
        #tv.grid(row=1, column=0, columnspan=2, sticky='nsew')
        
        
        
        dc_iid = tv.insert("", tkinter.END, text="HWSystem", values=['Top', 'Top level'], open=True)
        
        self.map_id_obj[dc_iid] = self.sys
    
        self.populateTree(tv, dc_iid, self.sys)
        
        print(tv.cget("displaycolumns"))
        tv.config(columns=[ 'instance', 'type'])
        tv.config(displaycolumns=[ 0, 1])
        tv.heading('instance', text='Instance name')
        tv.heading("type", text="Type")
        tv.heading("#0", text="Name")
        tv.config(selectmode=tkinter.BROWSE)
        #tv.tag_configure("ally", background="green")
        #tv.tag_bind("char", "<Double-Button-1>", event)
        
        #tv.config(style="Prolepsis.Treeview")
        tv["style"] = "Prolepsis.Treeview"
        #tv.config(bg="light yellow")
        print(ttk.Style().lookup("Prolepsis.Treeview", "background"))
        



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 12:40:41 2022

@author: dcr
"""

from .bitwise import *
from .arithmetic import *
from .arithmetic_fp import *
from .arithmetic_fxp import *
from .relational import *
from .simulation import *
from .clock import *
from .. import *



        
class Logic:
    """
    Base class for all logic circuits
    """
    
    def __init__(self, parent, instanceName:str):
        self.parent = parent
        self.name = instanceName
        
        if (not(parent is None)):
            if not(isinstance(parent, Logic)):
                raise Exception('parent object must be a Logic object not a {}'.format(type(parent)))
                
            # Add this object as a parent's child
            if (instanceName in parent.children.keys()):
                raise Exception('there is already a child named {} in {}'.format(instanceName, parent.getFullPath() ))

            parent.children[instanceName] = self

        self.inPorts = []   # in ports are a sorted list
        self.outPorts = []  # out ports are a sorted list
        self.sources = []   # list of InterfaceSource objects
        self.sinks = []     # List of InterfaceSink objects
    
        self.children = {}          # children are keyed by name
        self.clockDriver = None     # every circuit has a clock driver, if None it is inherited from parent
        
        self._wires = {}        # dictionary of wires created by the object
        
    def addIn(self, name , wire):
        port = InPort(self, name, wire)
        self.inPorts.append(port)
        return wire;
    
    def getInPortByName(self, name):
        for port in self.inPorts:
            if (port.name == name):
                return port
        return None

    def getOutPortByName(self, name):
        for port in self.outPorts:
            if (port.name == name):
                return port
        return None
        
    def reconnectIn(self, name, wire):
        # there is already a port with this name, but we just one to change the wire assigned
        port = self.getInPortByName(name)
        port.wire = wire
        
        
    def addOut(self, name, wire):
        port = OutPort(self, name, wire)
        self.outPorts.append(port)
        return wire;

    def addInterfaceSource(self, name:str, interface):
        """
        Adds the source ports of the interface to the 

        Parameters
        ----------
        Interface : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ports = []
        if (len(name) > 0):
            name = name + '_'
            
        for obj in interface.sourceToSink:
            wire = obj[1]
            portname = name + obj[0]
            port = OutPort(self, portname, wire)
            self.outPorts.append(port)
            ports.append(port)
        
        for obj in interface.sinkToSource:
            wire = obj[1]
            portname = name + obj[0]
            port = InPort(self, portname, wire)
            self.inPorts.append(port)
            ports.append(port)

        self.sources.append(InterfaceSource(name, ports))
            
        return interface;
            
    def addInterfaceSink(self, name:str, interface):
        """
        Adds the source ports of the interface to the 

        Parameters
        ----------
        Interface : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ports = []
        if (len(name) > 0):
            name = name + '_'
            
        for obj in interface.sourceToSink:
            wire = obj[1]
            portname = name + obj[0]
            port = InPort(self, portname, wire)
            self.inPorts.append(port)
            ports.append(port)
        for obj in interface.sinkToSource:
            wire = obj[1]
            portname = name + obj[0]
            port = OutPort(self, portname, wire)
            self.outPorts.append(port)
            ports.append(port)
            
        self.sinks.append(InterfaceSink(name, ports))
        
        return interface;
        
    def appendWire(self, wire):
        #print('checking if ', wire.name, 'is already in', [x for x in self._wires.keys()])
        if (wire.name in self._wires.keys()):
            raise Exception('a wire named {} already exist'.format(wire.name))
            
        self._wires[wire.name] = wire
        
    def wire(self, name, width=1):
        return Wire(self, name, width);
    
    def wires(self, name:str, num:int, width:int) -> list:
        """
        Creates a number of wires

        Parameters
        ----------
        name : str
            prefix of the wire names, it will be suffixed by _0, _1, etc.
        num : int
            numer of wires to create.
        width : int
            width of the wires to create.

        Returns
        -------
        list
            a list with the created wires.

        """
        ret = []
        for i in range(num):
            ret.append(self.wire('{}_{}'.format(name,i), width))
        return ret;
    
    def allLeaves(self):
        acum = []
        if (len(self.children) > 0):
            for obj in self.children.values():
                acum = acum + obj.allLeaves()
        else:
            acum = acum + [self];
            
        return acum
    
    def isPropagatable(self):
        return has_method(self, "propagate")
        
    def isClockable(self):
        return has_method(self, "clock")
    
    
    # def getHierarchy(self):
    #     str = ''
        
    #     if (self.parent != None):
    #         str = self.parent.getHierarchy()

    #     str = str + '/' + type(self).__name__  
        
    #     return str;
    
    def isPrimitive(self):
        return self.isClockable() or self.isPropagatable();
    
    def isStructural(self):
        # structural circuits have children
        return (len(self.children.values()) > 0)
        
    def getFullPath(self, withselfnames=True)->str:
        """
        Gets the full of a hierarchy element

        Returns
        -------
        str
            DESCRIPTION.

        """
        str = ''
        
        if (self.parent != None):
            str = self.parent.getFullPath(withselfnames=withselfnames)
            
        str = str + '/' + type(self).__name__ 
        if (withselfnames):
            str += '[' + self.name + ']' 
        
        return str

    
    def getFromFullPath(self, path:str):
        """
        Gets a Logic from the full path

        Parameters
        ----------
        path : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        pos0 = path.find('/', 0)
        pos1 = path.find('/', pos0+1)
        
        if (pos1 >=0):
            currentname = path[pos0+1:pos1]
        else:
            currentname = path[pos0+1:]
            
            
        pos2 = path.find('/', pos1+1)
        
        if (pos2 >=0):
            childname = path[pos1+1:pos2]
        else:
            childname = path[pos1+1:]
            
        parts = childname.split('[')
        childType = parts[0]
        parts2 = parts[1].split(']')
        childInstance = parts2[0]
        
        for child in self.children:
            if  ( type(child).__name__ == childType and child.name == childInstance):
                objFound = child
        
        if (pos2 >= 0):
            return objFound.getFromFullPath( path[pos1:])
        else:
            return objFound;
        
class Wire:
    """
    Wires in py4hw connect one source with one (or more) sinks
    """
    
    # this is the list of prepared wires, 
    prepared = []
    
    def __init__(self, parent, name : str, width: int = 1 ):
        # the following should not be necessary if python checks types
        assert(isinstance(name, str))
        assert(isinstance(width, int))
        
        self.parent = parent
        self.name = name
        self.width = width
        self.value = 0 # should be None      # reset state
        self.sinks = []
        self.source = None
        parent.appendWire(self)
        
    def getFullPath(self)->str:
        return self.parent.getFullPath() + '[{}]'.format(self.name)
    
    def getWidth(self) -> int:
        return self.width
    
    def put(self, val:int):
        mask = (1<<self.width) -1
        self.value = val & mask

    def prepare(self, val:int):
        mask = (1<<self.width) -1
        self.next = val & mask
        Wire.prepared.append(self)
        
    def settle(self):
        self.value = self.next
        
    def get(self) -> int:
        return self.value
    
    def setSource(self, source):
        if (self.source != None):
            raise Exception('Source of wire {} already connected to {}'.format(self.getFullPath(), self.source.parent.getFullPath()))

        self.source = source
        
        
    def getSource(self):
        """
        Returns the OutPort that drives this wire

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.source
    
    def addSink(self, sink):
        """
        Adds a sink to the wire

        Parameters
        ----------
        sink : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.sinks.append(sink)
        
    def getSinks(self):
        """
        Gets all the sinks from a wire

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.sinks;
    

    def settleAll():
        """
        Settles all pending wires that were changed by clock methods

        Returns
        -------
        None.

        """
        for w in Wire.prepared:
            w.settle()
            
        # empty list
        Wire.prepared = []
    
    def rename(self, newname):
        del self.parent._wires[self.name]
        self.name = newname
        self.parent.appendWire(self)
        
    def reparent(self, newparent):
        del self.parent._wires[self.name]
        self.parent = newparent
        newparent.appendWire(self)

    def reparentAndRename(self, newparent, newname):
        del self.parent._wires[self.name]
        self.name = newname
        self.parent = newparent
        newparent.appendWire(self)

class InPort:
    """
    An Input port
    """
    
    def __init__(self, parent:Logic, name, wire:Wire):
        """
        Creates an in port to the cell.
        The cell will be registered as sink to the wire only if it is
        a leaf (a primitive) in the hierarchy

        Parameters
        ----------
        parent : Logic
            DESCRIPTION.
        name : TYPE
            DESCRIPTION.
        wire : Wire
            DESCRIPTION.

        Returns
        -------
        None.

        """
        #print('in port')
        self.name = name
        self.parent = parent
        self.wire = wire

        if (parent.isPrimitive()):
            wire.addSink(self)
            
    def getFullPath(self):
        return self.parent.getFullPath() + '[{}]'.format(self.name)
    
class OutPort:
    """
    An output port
    """
    def __init__(self, parent:Logic, name:str, wire:Wire):
        """
        Creates an out port to the cell.
        The cell will be registered as source of the wire only if it
        is a leaf (a primitive) in the hierarchy 

        Parameters
        ----------
        parent : Logic
            Parent Cell of the port
        name : str
            name of the port.
        wire : Wire
            Wire associated with the port

        Returns
        -------
        None.

        """
        #print('out port')
        self.name = name
        self.parent = parent
        self.wire = wire
        
        if (parent.isPrimitive()):
            wire.setSource(self)

    def getFullPath(self):
        return self.parent.getFullPath() + '[{}]'.format(self.name)

class InterfaceSource:
    def __init__(self, name:str, ports):
        self.name = name
        self.ports = ports
    
    
    
class InterfaceSink:
    def __init__(self, name:str, ports):
        self.name = name
        self.ports = ports
    
class HWSystem(Logic):
    """
    A Hardware system is the top level entity of all designs 
    """
    def __init__(self, clock_driver=None, name=None):
        super().__init__(None, "HWSystem")
        self.simulator = None
        if (clock_driver is None):
            clock_driver = ClockDriver('clk50', 50E6, 0, wire=self.wire('clk50')) # we create a 50MHz clock driver by default
        if not(name is None):
            self.name = name
        self.clockDriver = clock_driver
        
        
    def getSimulator(self):
        """
        Returns the singleton simulator instance

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        from .simulation import Simulator

        if (self.simulator == None):
            self.simulator = Simulator(self)
        else: 
            self.simulator.topologicalSort() # Updates existing simulator
            
        return self.simulator
        
    
class ClockDriver():
    """
    A clock driver
    """
    
    def __init__(self, name:str, freq=50E6, phaseOffset=0, base=None, enable=None, wire=None):
        """
        Creates a clock driver

        Parameters
        ----------
        name : str
            Name of the clock driver.
        freq : number, optional
            frequency (in Hz) of the clock .
        phaseOffset : number, optional
            phase offset (in %) of the positive edge of the clock
        base : ClockDriver, optional
            ClockDriver that this clock driver is based on. This is useful for gated clocks.
        enable : Wire, optional
            Wire controling the gating of the clock. When enable is 0, the clock
            will be gated
        wire : Wire, optional
            Wire that contains the clock

        Returns
        -------
        None.

        """
        self.name = name
        self.base = base
        
        if (base is None):
            self.freq = freq
            self.phaseOffset = phaseOffset
        else:
            self.freq = base.freq;
            self.phaseOffset = base.phaseOffset
            
        self.enable = enable
        self.wire = wire

def has_method(o, name):
    return callable(getattr(o, name, None))


class Interface:
    """
    An interface is a collection of wires that connect
    a source to a sink
    """
    def __init__(self, parent:Logic, name:str):
        self.name = name
        self.parent = parent
        self.sourceToSink = []
        self.sinkToSource = []
    
    def addSourceToSink(self, name:str, width:int):
        """
        Adds a source wire to the definition of the interface

        Parameters
        ----------
        name : str
            DESCRIPTION.
        width : int
            DESCRIPTION.

        Returns
        -------
        None.

        """
        w = self.parent.wire(self.name + "_" + name, width)
        self.sourceToSink.append([name, w])
        return w;
    
    def getSourceToSink(self, name):
        if len(self.sourceToSink) == 0:
            raise Exception('No source-to-sink elements in this interface')
            
        for p in self.sourceToSink:
            if (p[0] == name):
                return p[1]
            
        raise Exception('SourceToSink {} not found'.format(name))
    
    def getSinkToSource(self, name):
        if len(self.sinkToSource) == 0:
            raise Exception('No sink-to-source elements in this interface')
            
        for p in self.sinkToSource:
            if (p[0] == name):
                return p[1]
                
        raise Exception('SinkToSource {} not found'.format(name))
        
    def addSinkToSource(self, name:str, width:int):
        """
        Adds a source wire to the definition of the interface

        Parameters
        ----------
        name : str
            DESCRIPTION.
        width : int
            DESCRIPTION.

        Returns
        -------
        None.

        """
        w = self.parent.wire(self.name + "_" + name, width)
        self.sinkToSource.append([name, w])
        return w;
        

def disconnectWireFromLogicObject(w:Wire, obj:Logic):
    """
    Disconnects a wire from a logic object

    Parameters
    ----------
    w : Wire
        The wire to disconnect from the object.
    obj : Logic
        The object to disconnect from the wire.

    Returns
    -------
    None.

    """
    if (w.source in obj.outPorts):
        # wire is connected to an output port of the obj
        port = w.source
        w.source = None
        port.wire = None
        return 

    for sink in w.sinks:
        if (sink in obj.inPorts):
            # wire is connected to an input port of the obj
            w.sinks.remove(sink)
            sink.wire = None
            return
        
    raise Exception('wire and object are not connected')
    
def getObjectClockDriver(obj:Logic) -> ClockDriver:
    """
    Returns the clock driver of an object.
    @todo consider multiple clock drivers scenario

    Parameters
    ----------
    obj : Logic
        Object of interest.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    ClockDriver
        The ClockDriver of the object.

    """
    
    if (obj.clockDriver != None):
        return obj.clockDriver
    if (obj.parent == None):
        raise Exception('No clock driver at top level')
    else:
        return getObjectClockDriver(obj.parent)


# -*- coding: utf-8 -*-
"""
Created on Tue Aug  9 06:15:25 2022

Arithmetic Floating-Point blocks

@author: dcr
"""
from .. import *
from .bitwise import *
from .arithmetic import *
from deprecated import deprecated


class _FP_parts(Logic):
    def __init__(self, parent:Logic, name:str, a:Wire, s:Wire, e:Wire, m:Wire, isZero:Wire, fixExponent=True):
        super().__init__(parent, name)

        from ..helper import LogicHelper
    
        assert(s.getWidth() == 1)
        assert(e.getWidth() == 8)
        assert(m.getWidth() == 24)
        
        self.addIn('a', a)
        self.addOut('s', s)
        self.addOut('e', e)
        self.addOut('m', m)
        self.addOut('isZero', isZero)

        pre_e = self.wire('pre_e', 8)
        pre_m = self.wire('pre_m', 23)
        
        Bit(self, 's',a, 31, s)
        Range(self, 'pre_e', a, 30, 23, pre_e)
        Range(self, 'm', a, 22, 0, pre_m)
        
        one = self.wire('one', 1)
        Constant(self, 'one', 1, one)

        ConcatenateMSBF(self, 'ma', [one, pre_m], m)

        g = LogicHelper(self)

        if (fixExponent):
            Sub(self, 'e', pre_e , g.hw_constant(8, 127), e)
        else:
            Buf(self, 'e', pre_e, e)

        Buf(self, 'isZero', g.hw_or2(g.hw_equal_constant(pre_e, 0),
                                     g.hw_equal_constant(pre_m, 0)), isZero)        
        
class FPAdder_SP(Logic):
    # This design is a very basic one, inspired in algorithm described in 
    # the 2.1 section of the paper 
    # Oberman, Stuart F., and Michael J. Flynn. "A variable latency pipelined 
    # floating-point adder." In European Conference on Parallel Processing, 
    # pp. 183-192. Springer, Berlin, Heidelberg, 1996.
    # https://link.springer.com/content/pdf/10.1007/bfb0024701.pdf
    def __init__(self, parent:Logic, name:str, a:Wire, b:Wire, r:Wire):
        super().__init__(parent, name)
        
        
        # This is really cumbersome
        from ..helper import LogicHelper
                    
        a = self.addIn('a', a)
        b = self.addIn('b', b)
        r = self.addOut('r', r)
        
        igt = self.wire('igt')
        ieq = self.wire('ieq')
        ilt = self.wire('ilt')

        FPComparator_SP(self, 'cmp', a, b, igt, ieq, ilt, absolute=True)
        
        a2 = self.wire('a2', a.getWidth())
        b2 = self.wire('b2', b.getWidth())
        
        Swap(self, 'swap', a, b, ilt, a2, b2)
        
        a = a2
        b = b2
        
        sa = self.wire('sa')
        sb = self.wire('sb')
        ea = self.wire('ea', 8)
        eb = self.wire('eb', 8)
        ma = self.wire('ma', 24)
        mb = self.wire('mb', 24)
        isZeroa = self.wire('isZeroa')
        isZerob = self.wire('isZerob')
        
        _FP_parts(self, 'parts_a', a, sa, ea, ma, isZeroa, fixExponent=False)
        _FP_parts(self, 'parts_b', b, sb, eb, mb, isZerob, fixExponent=False)
        
        # Maximum possible shifting is 23 bits (of the mantisa), so
        # it is enough with 5 bits for ediff
        # Also we know ediff will be always positive

        ediff = self.wire('ediff', 5)
        Sub(self, 'ediff', ea, eb, ediff)
        
        mb3 = self.wire('mb3', mb.getWidth())
        
        ShiftRight(self, 'preshift', mb, ediff, mb3)
        
        mb = mb3
        
        s_eq = self.wire('s_eq')
        Equal(self, 's_eq', sa, sb, s_eq)
        
        m_a_plus_b = self.wire('m_a_plus_b', ma.getWidth()+1)
        m_a_minus_b = self.wire('m_a_minus_b', ma.getWidth()+1)
        #m_b_minus_a = self.wire('m_b_minus_a', ma.getWidth()+1)
        
        Add(self, 'm_a_plus_b', ma, mb, m_a_plus_b)
        Sub(self, 'm_a_minus_b', ma, mb, m_a_minus_b)
        #Sub(self, 'm_b_minus_a', mb, ma, m_b_minus_a)
        
        # this is unnecessary since a > b
        #samb = self.wire('samb')
        #sbma = self.wire('sbma')
        
        #Sign(self, 'samb', m_a_minus_b, samb)
        #Sign(self, 'sbma', m_b_minus_a, sbma)
        
        
        mr = self.wire('mr', m_a_plus_b.getWidth())
        sr = self.wire('sr')
        
        g = LogicHelper(self)
        #sel_apb = g.hw_buf(s_eq)
        sel_amb = g.hw_xor2(sa, sb)
        #sel_bma = g.hw_and2(g.hw_not(sa), sb)
        
        Mux2(self, 'select_mr', sel_amb, m_a_plus_b, m_a_minus_b, mr)
        Buf(self, 'sr', sa, sr)
        
        
        clz = self.wire('clz', 5)
        clzz = self.wire('clzz')
        
        CountLeadingZeros(self, 'clz', mr, clz, clzz)
        
        mr2 = self.wire('mr2', mr.getWidth())
        ShiftLeft(self, 'shift_left', mr, clz, mr2)
        #Select(self, 'select_mr', [sel_apb, sel_amb, sel_bma], [m_a_plus_b, m_a_minus_b, m_b_minus_a], mr)
        #Select(self, 'select_sr', [sel_apb, sel_amb, sel_bma], [sa, sb, sa], sr)

        # Compute the result exponent
        pre_er = self.wire('pre_er', 8)
        er = self.wire('er', 8)
                
        Sub(self, 'pre_er', ea, clz, pre_er)
        Add(self, 'er', pre_er, g.hw_constant(8, 1), er)
        
        # Detect round up bit
        round_up = self.wire('round_up')
        Bit(self, 'round_up', mr2, 1, round_up)
        
        # Add 1 (actually b10) for rounding up mr2
        round_one = self.wire('round_one', 25)
        round_result = self.wire('round_result', 26)
        round_result_msb = self.wire('round_result_msb')
        Constant(self, 'round_one', 1, round_one)
        Add(self, 'round_result', mr2, round_one, round_result)
        Bit(self, 'round_result_msb', round_result, 25, round_result_msb)
        
        # invert result
        #inverted = self.wire('inverted')
        #mr2 = self.wire('mr2', mr.getWidth())
        #Abs(self, 'abs', mr, mr2, inverted)        
        #mr = mr2
        
        # invert the sign if necessary
        #sr = g.hw_xor2(inverted, sr)
        
        # shrink wire, we take from 23 to 1, as we assume that 24 is 1 
        mr3 = self.wire('mr3', 23)        
        Range(self, 'mr3', mr2, 23, 1, mr3)
                
        ConcatenateMSBF(self, 'final_r', [sr, er, mr3], r)
        
        
        
class FPtoInt_SP(Logic):
    # This design is inspired in algorithm described in 
    # section 9.2.1 of the book 
    # Computer Principles and design in Verilog
    def __init__(self, parent:Logic, name:str, a:Wire, r:Wire, p_lost:Wire, denorm:Wire, invalid:Wire):
        super().__init__(parent, name)

        # This is really cumbersome
        from ..helper import LogicHelper

        self.addIn('a', a)
        self.addOut('r', r)
        self.addOut('p_lost', r)
        self.addOut('denorm', denorm)
        self.addOut('invalid', invalid)

        g = LogicHelper(self)
        e = g.hw_range(a, 30, 23)
        m = g.hw_range(a,22,0)
        hidden_bit = g.hw_not_equal_constant(e, 0)        
        frac_is_not_0 = g.hw_not_equal_constant(m, 0)

        is_denorm = g.hw_and2(g.hw_not(hidden_bit), frac_is_not_0)
        is_zero = g.hw_and2(g.hw_not(hidden_bit), g.hw_not(frac_is_not_0))
        sign = g.hw_bit(a,31)
        
        zero_tail = g.hw_constant(23, 0)
        real_m = g.hw_concatenate_msbf([hidden_bit, m]) 
        #frac0 = g.hw_concatenate_msbf([hidden_bit, m, zero_tail]) # 56 bits
        
        # e2 = 32 + (e - 127)
        real_e = g.hw_sub(e, g.hw_constant(8, 127))
        
        sign_real_e = g.hw_sign(real_e)
        
        k31 = g.hw_constant(real_e.getWidth(), 30)
        #too_big = self.wire('too_big')
        e_gt_31 = self.wire('e_gt_31')
        e_eq_31 = self.wire('e_eq_31')
        e_lt_31 = self.wire('e_lt_31')
        Comparator(self, 'too_big', real_e, k31, e_gt_31, e_eq_31, e_lt_31)
        too_big = g.hw_or2(g.hw_and2(e_gt_31, g.hw_not(sign_real_e)),
                           g.hw_and2(sign, e_eq_31))
        
        # if sign is 1, this means that the number is < 0
        
        # 32 is the range of integers, 23 the range of the fractional part of m        
        expanded_m = self.wire('expanded_m', 32+24) 
        ShiftLeft(self, 'expanded_m', real_m, real_e, expanded_m)

        integer_m = g.hw_range(expanded_m, 32+23-1, 23)
        fractional_m = g.hw_range(expanded_m, 22, 0)
        franctional_m_not_zero = g.hw_not_equal_constant(fractional_m, 0)
        
        positive_m = integer_m
        negative_m = g.hw_neg(integer_m)
        
        final_m = g.hw_mux2(sign, positive_m, negative_m)
        #wire [8:0] shift_right_bits = 9’d158 - {1’b0,a[30:23]}; // 127 + 31
        
        # for denorm values
        select_denorm = is_denorm
        p_lost_denorm = g.hw_constant(1,1)
        invalid_denorm = g.hw_constant(1,0)
        d_denorm = g.hw_constant(32, 0)
        
        # small values
        select_small = sign_real_e
        p_lost_small = frac_is_not_0 
        invalid_small = g.hw_constant(1,0)
        d_small = g.hw_constant(32, 0)
        
        # default
        select_default = g.hw_or2(g.hw_not(is_denorm), g.hw_not(select_small))
        p_lost_default = franctional_m_not_zero
        invalid_default = too_big
        d_default = final_m

        Buf(self, 'denorm', is_denorm, denorm)
        Select(self, 'select_p_lost', [select_denorm, select_small, select_default], 
               [p_lost_denorm, p_lost_small, p_lost_default], p_lost)
        Select(self, 'select_invalid', [select_denorm, select_small, select_default],
               [invalid_denorm, invalid_small, invalid_default], invalid)
        Select(self, 'select_d', [select_denorm, select_small, select_default], 
               [d_denorm, d_small, d_default], r)


class InttoFP_SP(Logic):
    # This design is inspired in algorithm described in 
    # section 9.2.2 of the book 
    # Computer Principles and design in Verilog
    def __init__(self, parent:Logic, name:str, a:Wire, r:Wire, p_lost:Wire):
        super().__init__(parent, name)

        # This is really cumbersome
        from ..helper import LogicHelper

        assert(a.getWidth() == 32)        
        assert(r.getWidth() == 32)        

        self.addIn('a', a)
        self.addOut('r', r)
        self.addOut('p_lost', p_lost)

        g = LogicHelper(self)
        
        sign = self.wire('sign')
        f5 = self.wire('f5', a.getWidth())
        
        Abs(self, 'f5', a, f5, inverted=sign)

        clz = self.wire('clz', 5)
        is_zero = self.wire('is_zero')
        
        CountLeadingZeros(self, 'clz', f5, clz, is_zero)

        shifted = self.wire('shited', 32)
        ShiftLeft(self, 'shift', f5, clz, shifted)
        
        # we skip the first 1, as it is implicit
        fraction = g.hw_range(shifted, 30, 30-23+1)

        pre_p_lost = g.hw_not_equal_constant(g.hw_range(shifted, 30-23, 0), 0)
        Buf(self, 'p_lost', pre_p_lost, p_lost)

        exponent = g.hw_sub(g.hw_constant(8, 127+31), clz)

        #is_zero = g.hw_equal_constant(a, 0)
        pre_r = self.wire('pre_r', 32)
        ConcatenateMSBF(self, 'pre_r', [sign, exponent, fraction], pre_r)
        Mux2(self, 'r', is_zero, pre_r, g.hw_constant(32, 0), r)
        
        
class FPMult_SP(Logic):
    # 
    def __init__(self, parent:Logic, name:str, a:Wire, b:Wire, r:Wire):
        super().__init__(parent, name)
        
        self.addIn('a', a)
        self.addIn('b', b)
        self.addOut('r', r)
        
        sa = self.wire('sa')
        sb = self.wire('sb')
        ea = self.wire('ea', 8)
        eb = self.wire('eb', 8)
        ma = self.wire('ma', 24)
        mb = self.wire('mb', 24)
        isZeroa = self.wire('isZeroa')
        isZerob = self.wire('isZerob')
        
        _FP_parts(self, 'pa', a, sa, ea, ma, isZeroa, fixExponent=False)
        _FP_parts(self, 'pb', b, sb, eb, mb, isZerob, fixExponent=False)

        from ..helper import LogicHelper        
        g = LogicHelper(self)
        
        # a * b = sa * sb * 2^ea * 2^eb * ma * mb
        # a * b = (sa*sb) * 2 ^(ea+eb) + (ma*mb)
        
        isZeror = g.hw_or2(isZeroa, isZerob)
        
        sr = g.hw_xor2(sa, sb)
        
        pre_mr = self.wire('pre_mr', ma.getWidth() + mb.getWidth())
        Mul(self, 'mult', ma, mb, pre_mr)
        
        pre_er = self.wire('pre_er', 9)
        Add(self, 'add', ea, eb, pre_er)
        pre_er2 = self.wire('pre_er2', 8)
        pre_er3 = self.wire('pre_er3', 8)
        Sub(self, 'pre_er2', pre_er, g.hw_constant(9, 127-1), pre_er2)
        Sub(self, 'pre_er3', pre_er2, g.hw_constant(9,1), pre_er3)
        
        # result of ma*mb must be 0 <= mr < 4, it seems the first bit will be always 1
        select_mr = g.hw_bit(pre_mr, 47)
        pre_mr2 = g.hw_range(pre_mr, 46, 24)
        pre_mr3 = g.hw_range(pre_mr, 45, 23)
        
        pre_mr4 = g.hw_mux2(select_mr, pre_mr3, pre_mr2)
        pre_er4 = g.hw_mux2(select_mr, pre_er3, pre_er2)
        
        ConcatenateMSBF(self, 'r', [sr, pre_er4, pre_mr4], r)


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 27 14:06:19 2022

@author: dcr
"""
from .. import *

class GatedClock(Logic):
    
    def __init__(self, parent:Logic, name:str, enin:Wire , enout:Wire, drv:ClockDriver):
        super().__init__(parent, name)
        
        self.enin = self.addIn('enin', enin)
        self.enout = self.addOut('enout', enout)
        self.drv = drv
        
    def propagate(self):
        # this is fake, just to make simulator work
        self.enout.put(self.enin.get())
        
        
class ClockDivider(Logic):
    def __init__(self, parent, name, freq_in, freq_out, clkout):
        from py4hw.helper import LogicHelper
        from py4hw.logic.arithmetic import ModuloCounter
        from py4hw.logic.storage import TReg
        import math
        super().__init__(parent, name)
            
        req_freq_out = freq_out
        hlp = LogicHelper(self)
        
        self.addOut('clkout', clkout)
        
        n = freq_in / (2*freq_out)
        
        qw = int(math.log2(n)) + 1
        
        n = int(n)
        
        freq_out = freq_in / (2*n)
        
        #print('n', n, 'qw', qw, 'freq_out', freq_out)
        if (freq_out != req_freq_out):
            print('WARNING: Real Output Frequency:', freq_out, 'required:', req_freq_out)
        
        assert(qw > 0)
        q = self.wire('q', qw)
        t = self.wire('t')
        inc = hlp.hw_constant(1,1)
        reset = hlp.hw_constant(1,0)
        dut = ModuloCounter(self, 'count', mod=n, inc=inc, reset=reset, q=q, carryout=t)
    
        TReg(self, 'clkout', t, e=inc, q=clkout)
        


# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 12:53:47 2022

@author: dcr
"""
from .. import *
from .bitwise import *

from deprecated import deprecated


class Add(Logic):
    """
    Combinational Arithmetic Add
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire, ci=None, co=None):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)
        
        if (ci is None):
            self.ci = None
        else:
            self.ci = self.addIn('ci', ci)
            
        if (co is None):
            self.co = None
        else:
            self.co = self.addOut('co', co)

    def propagate(self):
        if (not(self.ci is None)):
            vci = self.ci.get()
        else:
            vci = 0

        vr = self.a.get() + self.b.get() + vci
        
        if (not(self.co is None)):
            self.co.put(vr >> self.r.getWidth())
            
        self.r.put(vr)

class Abs(Logic):
    """
    Absolute value
    """
    def __init__(self, parent, name: str, a: Wire, r: Wire, inverted:Wire=None):
        """
        Creates ans absolute value circuit r = abs(a), inverted=sign(a)

        Parameters
        ----------
        parent : TYPE
            parent.
        name : str
            name.
        a : Wire
            input.
        r : Wire
            absolute value of a.
        inverted : Wire, optional
            indicates wether a was negated. The default is None.

        Returns
        -------
        None.

        """
        super().__init__(parent, name)
        a = self.addIn("a", a)
        r = self.addOut("r", r)

        if not(inverted is None):
            s = self.addOut('inverted', inverted)
        else:
            s = self.wire('sign')

        zero = self.wire('zero', r.getWidth())
        neg = self.wire('neg', a.getWidth())
        Constant(self, 'zero', 0, zero)
        Sign(self, 'sign', a, s)
        Sub(self, 'sub', zero, a, neg)
        Mux2(self, 'mux', s, a, neg, r)


class Neg(Logic):
    def __init__(self, parent, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)

        self.addIn("a", a)
        self.addOut("r", r)

        zero = self.wire('zero', r.getWidth())
        Constant(self, 'zero', 0, zero)    
        Sub(self, 'sub', zero, a, r)

class Sign(Logic):
    """
    Sign test.
    r = 0 if a >= 0 (positive)
    t = 1 if a < 0 (negative)
    """

    def __init__(self, parent, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.r = self.addOut("r", r)

        Bit(self, "signBit", a, a.getWidth() - 1, r)
        

class SignExtend(Logic):
    """
    Behaviouraly modeled sign extend
    """
    def __init__(self, parent: Logic, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)

    def propagate(self):
        value = self.a.get()
        hb = value >> (self.a.getWidth() - 1)
        for i in range(self.a.getWidth(), self.r.getWidth()):
            value = value | (hb << i)

        self.r.put(value)


class ZeroExtend(Logic):
    """
    Behaviouraly modeled zero extend
    """
    def __init__(self, parent: Logic, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)

    def propagate(self):
        value = self.a.get()
        self.r.put(value)





class Mul(Logic):
    """
    Combinational Arithmetic Multiplier
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.a.get() * self.b.get())

class SignedMul(Logic):
    """
    Arithmetic Signed Multiplier
    """
    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        from ..helper import IntegerHelper    
        
        sa = IntegerHelper.c2_to_signed(self.a.get(), self.a.getWidth())
        sb = IntegerHelper.c2_to_signed(self.b.get(), self.b.getWidth())
        mask = (1 << self.r.getWidth()) - 1
        newValue = (sa * sb) & mask
        self.r.put(newValue)
        
class Div(Logic):
    """
    Combinational Arithmetic Divider
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.a.get() // self.b.get())

class Mod(Logic):
    """
    Combinational Arithmetic Modulo (reminder of div)
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.a.get() % self.b.get())


class SignedDiv(Logic):
    """
    Combinational Arithmetic Divider
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

        abs_a = self.wire('abs_a', a.getWidth())
        abs_b = self.wire('abs_b', b.getWidth())
        
        Abs(self, 'abs_a', a, abs_a)
        Abs(self, 'abs_b', b, abs_b)
        
        sign_a = self.wire('sign_a')
        sign_b = self.wire('sign_b')
        
        Sign(self, 'sign_a', a, sign_a)
        Sign(self, 'sign_b', b, sign_b)
       
        q = self.wire('q', r.getWidth())        
        neg_q = self.wire('neg_q', r.getWidth())
        
        Div(self, 'div', abs_a, abs_b, q)
        Neg(self, 'neg', q, neg_q)
        
        sign_r = self.wire('sign_r')
        Xor2(self, 'sign_r', sign_a, sign_b, sign_r) 
        
        Mux2(self, 'r', sign_r, q, neg_q, r)        
        

class Sub(Logic):
    """
    Arithmetic Sub
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        mask = (1 << self.r.getWidth()) - 1
        newValue = (self.a.get() - self.b.get()) & mask
        self.r.put(newValue)


class SignedSub(Logic):
    """
    Arithmetic Sub
    """
    

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        from ..helper import IntegerHelper    
        
        sa = IntegerHelper.c2_to_signed(self.a.get(), self.a.getWidth())
        sb = IntegerHelper.c2_to_signed(self.b.get(), self.b.getWidth())
        mask = (1 << self.r.getWidth()) - 1
        newValue = (sa - sb) & mask
        self.r.put(newValue)
        
class Counter(Logic):
    """
    Counts up to the value mod and returns to zero
    """
    def __init__(self, parent, name : str, reset:Wire , inc:Wire , q:Wire ):
        super().__init__(parent, name)
        
        from .bitwise import Constant
        from .bitwise import Or2
        from .bitwise import Mux2
        from .storage import Reg
        
        reset = self.addIn('reset', reset)
        inc = self.addIn('inc', inc)
        q = self.addOut('q', q)
    
        one = self.wire('one', q.getWidth())
        zero = self.wire('zero', q.getWidth())
        add = self.wire('add', q.getWidth())
        d = self.wire('d', q.getWidth())
        d1 = self.wire('d1', q.getWidth())
        e_add = self.wire('e_add', 1)
        
        Constant(self, 'one', 1, one)
        Constant(self, 'zero', 0, zero)
        
        Mux2(self, 'muxinc', inc, q, add, d1)
        Mux2(self, 'muxreset', reset, d1, zero, d)

        #py4hw.Select(self, 'select', [reset, inc], [zero, add], d)
        Or2(self, 'e_add', reset, inc, e_add)
        #py4hw.Mux(self, 'mux', )
        Add(self, 'add', q, one, add)
        Reg(self, 'reg', d, q, e_add)
        
class ModuloCounter(Logic):
    """
    Counts up to the value mod and returns to zero
    """
    def __init__(self, parent, name : str, mod : int, reset , inc , q , carryout):
        super().__init__(parent, name)
        
        from .bitwise import Constant
        from .bitwise import Or2
        from .bitwise import Mux2
        from .storage import Reg
        from .relational import EqualConstant
        
        reset = self.addIn('reset', reset)
        inc = self.addIn('inc', inc)
        q = self.addOut('q', q)
        carryout = self.addOut('carryout', carryout)
    
        one = self.wire('one', q.getWidth())
        zero = self.wire('zero', q.getWidth())
        add = self.wire('add', q.getWidth())
        d = self.wire('d', q.getWidth())
        d1 = self.wire('d1', q.getWidth())
        e_add = self.wire('e_add', 1)
        anyreset = self.wire('anyreset', 1)
        
        Constant(self, 'one', 1, one)
        Constant(self, 'zero', 0, zero)
        
        Or2(self, 'anyreset', reset, carryout, anyreset)
        Mux2(self, 'muxinc', inc, q, add, d1)
        Mux2(self, 'muxreset', anyreset,  d1,zero, d)

        #py4hw.Select(self, 'select', [reset, inc], [zero, add], d)
        Or2(self, 'e_add', reset, inc, e_add)
        #py4hw.Mux(self, 'mux', )
        Add(self, 'add', q, one, add)
        Reg(self, 'reg', d, q, enable=e_add)
        EqualConstant(self, 'eq{}'.format(mod-1), q, mod-1, carryout)
        
        
class ShiftRight(Logic):
    def __init__(self, parent:Logic, name:str, a, b, r):
        super().__init__(parent, name)

        a = self.addIn('a', a)
        b = self.addIn('b', b)
        r = self.addOut('r', r)
            
        wb = b.getWidth()
        
        num = int(math.pow(2, wb))
        ins = []
        
        for i in range(num):
            shifted = self.wire('shifted{}'.format(i), r.getWidth())
            ShiftRightConstant(self, 'shifted{}'.format(i), a, i, shifted)
            ins.append(shifted)
            
        Mux(self, 'mux', b, ins, r)

class ShiftLeft(Logic):
    def __init__(self, parent:Logic, name:str, a, b, r):
        super().__init__(parent, name)

        a = self.addIn('a', a)
        b = self.addIn('b', b)
        r = self.addOut('r', r)
            
        wb = b.getWidth()
        
        num = int(math.pow(2, wb))
        ins = []
        
        for i in range(num):
            shifted = self.wire('shifted{}'.format(i), r.getWidth())
            ShiftLeftConstant(self, 'shifted{}'.format(i), a, i, shifted)
            ins.append(shifted)
            
        Mux(self, 'mux', b, ins, r)
        
class BinaryToBCD(Logic):
    """
    Converts a binary number into BCD digits
    """
    
    def __init__(self, parent, name : str, a: Wire, r:Wire):
        from ..helper import LogicHelper    

        super().__init__(parent, name)
        
        a = self.addIn('a', a)
        r = self.addOut('r', r)
    
        hlp = LogicHelper(self)
        
        w = a.getWidth()
        assert(r.getWidth() % 4 == 0)
        digits = r.getWidth() // 4 # int(math.ceil(math.log10((2**w)-1)))
        print('Number of BCD digits:', digits)
        print('r width:', r.getWidth())
        
        assert(r.getWidth() >= (digits*4))
        
        ret = []
        v = a
        k10 = hlp.hw_constant(4, 10)
        
        for i in range(digits):
            rem = self.wire('mod{}'.format(i), 4)
            div = self.wire('div{}'.format(i), w)
            Mod(self, 'mod{}'.format(i), v, k10, rem)
            Div(self, 'div{}'.format(i), v, k10, div)
            ret.append(rem)
            v = div
            
        ConcatenateLSBF(self, 'r', ret, r)
        

class _FFunction(Logic):
    """
    F function described in the paper DOI: 10.1109/TVLSI.2008.2000458
    """
    def __init__(self, parent, name : str, a: list, r:Wire):
        super().__init__(parent, name)

        w = len(a)
        an = []
        
        for i in range(w):
            self.addIn('in{}'.format(i), a[i])
            ann = self.wire('an{}'.format(i))
            Not(self, 'an{}'.format(i), a[i], ann)
            an.append(ann)
            
        self.addOut('r', r)
        
        products = []
        notcount = 0
        idx = w-1
        negidx_start = w-2
        negidx_stop = w-2
        
        while (True):
            prodsig = [a[idx]]
            #print('positive:', idx, 'negative:', end='')
            
            for j in range(negidx_start, negidx_stop, -2):
                #print(j, end=',')
                prodsig.append(an[j])
                
            #print()
            prod = self.wire('prod{}'.format(idx))
            
            if (len(prodsig) > 1):
                And(self, 'and{}'.format(idx), prodsig, prod)
                products.append(prod)
            else:
                products.append(prodsig[0])
                
            idx -= 2
            negidx_stop -= 2
            
            if (idx < 0):
                break;
        
        if (len(products) > 1):
            Or(self, 'or', products, r)
        else:
            Buf(self, 'r', products[0], r)
        
class CountLeadingZeros(Logic):
    """
    Count leading zero bits
    We implement the design described in 
    Dimitrakopoulos, Giorgos, Kostas Galanopoulos, Christos Mavrokefalidis, 
    and Dimitris Nikolos. "Low-power leading-zero counting and anticipation 
    logic for high-speed floating point units." IEEE transactions on very large 
    scale integration (VLSI) systems 16, no. 7 (2008): 837-850.
    https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=4539802    
    DOI: 10.1109/TVLSI.2008.2000458
    """
    def __init__(self, parent, name : str, a: Wire, r:Wire, z:Wire):
        super().__init__(parent, name)
        
        a = self.addIn('a', a)
        r = self.addOut('r', r)
        z = self.addOut('z', z)

        aw = a.getWidth()
        rw = r.getWidth()
        
        r_intern_w = int(math.ceil(math.log2(aw)))
        if (r_intern_w > rw):
            raise Exception('r with too small for a input')
            
        # we must extend the input to a power of 2
        # if we do that, we should subtract the extra bits from the result
        a_intern_w = int(math.pow(2, r_intern_w))
        a_intern = self.wire('a_intern', a_intern_w)
        
        ZeroExtend(self, 'zexta', a, a_intern)
            
        r_intern = self.wire('r_intern', r_intern_w)
        
        r_preout = self.wire('r_preout', r.getWidth())
        
        # we support bigger than necessary outputs by automatically
        # zero extending
        ZeroExtend(self, 'zextr', r_intern, r_preout)
        
        # Work with individual a wires
        a_bits = self.wires('a', a_intern_w, 1)
        BitsLSBF(self, 'a_bits', a_intern, a_bits)
        
        # Work with indidual wires , and concatenate them into the r_intern
        r_bits = self.wires('r_intern', r_intern_w, 1)
        
        ConcatenateLSBF(self, 'concat', r_bits, r_intern)
        
        f_bits = a_bits
        
        #print('len f_bits:', len(f_bits))
        
        for i in range(r_intern_w):
            fvalue = self.wire('f{}'.format(i))
            _FFunction(self, 'f_{}'.format(i), f_bits, fvalue)
            Not(self, 'z_{}'.format(i), fvalue, r_bits[i])
            
            next_f_bits_w = len(f_bits)//2
            next_f_bits = self.wires('f{}'.format(i), next_f_bits_w, 1)
            
            #print('level', i, next_f_bits_w)
            for j in range(next_f_bits_w):

                Or2(self, 'o{}_{}'.format(i,j), f_bits[j*2], f_bits[j*2+1], next_f_bits[j])
            
            f_bits = next_f_bits
            
        # we should end with a single wire in f_bits
        Not(self, 'not_z', f_bits[0], z)
        
        allZeroK = self.wire('allZeroK', r.getWidth())
        Constant(self, 'allZeroK', a.getWidth(), allZeroK)

        if (a_intern_w > aw):
            r_preout2 = self.wire('preout2', rw)
            extra = self.wire('extra', rw)
            Constant(self, 'extra', a_intern_w-aw, extra)
            Sub(self, 'sub', r_preout, extra, r_preout2)
            r_preout = r_preout2

        Mux2(self, 'final', z, r_preout, allZeroK, r)
        


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  4 16:22:33 2023

@author: dcastel1
"""

import py4hw
from py4hw.base import *
from .astutils import * 

class Python2Structural:

    def __init__(self, obj:Logic, methodName:str):
        self.obj = obj
        self.methodName = methodName
        self.signals = {}
        self.indent = 0

    def generateStructural(self, instanceName):
        '''
        Transpile RTL style behavioural descriptions

        Returns
        -------
        str
            the equivalent RTL

        '''
        module = getMethod(self.obj, self.methodName)
    
        
        node = ReplaceBasic1().visit(module)        
        node = ReplaceBasic2().visit(node)        
        
        return HLSCircuit(self.obj.parent, instanceName, self.obj, node)
    
class HLSCircuit(py4hw.Logic):
    def __init__(self, parent, name, obj, node):
        super().__init__(parent, name)

        wires = {}
        wc = 0
        ic = 0
        
        # copy the reference circuit interface
        for inPort in obj.inPorts:
            ip = self.addIn(inPort.name, inPort.wire)
            wires[inPort.name] = ip
            
        for outPort in obj.outPorts:
            # @todo in the future we should disconnect behavioural and reuse
            # the wire, instead of creating a new wire
            
            ow = parent.wire('hls_'.format(outPort.name), outPort.wire.getWidth())
            op = self.addOut(outPort.name, ow)
            wires[outPort.name] = op
            
        for st in node.body:
            if (isinstance(st, RTLSynchronousAssignment)):
                left_reg = wires[st.left.name]
                right_reg = self.wire('w{}'.format(wc), left_reg.getWidth())
                wc += 1
                
                py4hw.Reg(self, 'i{}'.format(ic), d=right_reg, q=left_reg)
                ic += 1
                
                if (isinstance(st.right, RTLOperator)):
                    if (st.right.op == '+'):
                        left_add = wires[st.right.left.name]
                        right_add = wires[st.right.right.name]
                          
                        py4hw.Add(self, 'i{}'.format(ic), left_add, right_add, right_reg)
                        ic += 1
                    else:
                        raise Exception('operator {} not supported'.format(st.right.op))
                else:
                    raise Exception('right part type {} not supported'.format(st.right))
            else:
                raise Exception('statement {} not supported'.format(st))

class ReplaceBasic1(ast.NodeTransformer):
    
    def visit_Module(self, node):
        body = node.body[0].body
        body = RTLBody(body)
        node = ast.NodeTransformer.generic_visit(self, body)
        return node 
    
    
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        print('checking call', attr)
        if (attr == 'get'):
            #if isinstance(node.func.value, ast.Attribute):
            wirename = getAstName(node.func.value)
            #print('REPLACING GET ', wirename)
            ret =  RTLWire(wirename)
        elif (attr == 'prepare'):
            left = RTLWire(node.func.value.attr)
            ret = RTLSynchronousAssignment(left, node.args[0])        
        else:
            raise Exception('Function {} not supported'.format(attr))
            
        ret = ast.NodeTransformer.generic_visit(self, ret)
        
        return ret

class ReplaceBasic2(ast.NodeTransformer):
    
    
    def visit_BinOp(self, node):
        #print('replacing BinOp')
        node =  RTLOperator(node.left, node.op, node.right)
        node = ast.NodeTransformer.generic_visit(self, node)
        return node

    def visit_Attribute(self, node):
        return RTLWire(node.attr)

    def visit_Expr(self, node):        
        ret = node.value
        ret = ast.NodeTransformer.generic_visit(self, ret)
        return ret
    


class RTLSynchronousAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

class RTLBody(ast.AST):
    '''
    AST node to wrap body
    '''
    def __init__(self, body:list):
        self.body = body
        self._fields = tuple(['body', 'dummy'])

class RTLWire(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str):
        self.name = name
        self._fields = tuple(['name', 'dummy'])

    def toVerilog(self):
        return self.name


class RTLOperator(ast.AST):
    def __init__(self, left, op, right):
        if (isinstance(op, list)):
            # we only support single operators
            assert(len(op) == 1)
            op = op[0]
            
        self.left = left
        self.op = self.getOp(op)
        self.right = right
        self._fields = tuple(['left', 'op', 'right'])

    def getOp(self, operator):
        # translated an AST operator into verilog syntax
        if (isinstance(operator, ast.Add)):
            return '+'
        elif (isinstance(operator, ast.Eq)):
            return '=='
        else:
            raise Exception('operator {} not supported'.format(type(operator)))
